@@ -19,15 +19,17 @@
  */
 
 #include <stdio.h>
+#include <unistd.h>
 #include <getopt.h>
 #include <limits.h>
+#include <sys/types.h>
 
 #include "c.h"
 #include "fileutils.h"
 #include "nls.h"
 #include "xalloc.h"
-#include "proc/readproc.h"
-#include "proc/version.h" /* procps_version */
+
+#include <proc/pids.h>
 
 
 #define grow_size(x) do { \
@@ -55,6 +57,8 @@ static char *program = NULL;
 static int opt_single_shot    = 0;  /* -s */
 static int opt_scripts_too    = 0;  /* -x */
 static int opt_rootdir_check  = 0;  /* -c */
+static int opt_with_workers   = 0;  /* -w */
+static int opt_quiet          = 0;  /* -q */
 
 static char *pidof_root = NULL;
 
@@ -68,6 +72,8 @@ static int __attribute__ ((__noreturn__)) usage(int opt)
 	fputs(USAGE_OPTIONS, fp);
 	fputs(_(" -s, --single-shot         return one PID only\n"), fp);
 	fputs(_(" -c, --check-root          omit processes with different root\n"), fp);
+	fputs(_(" -q,                       quiet mode, only set the exit code\n"), fp);
+	fputs(_(" -w, --with-workers        show kernel workers too\n"), fp);
 	fputs(_(" -x                        also find shells running the named scripts\n"), fp);
 	fputs(_(" -o, --omit-pid <PID,...>  omit processes with PID\n"), fp);
 	fputs(_(" -S, --separator SEP       use SEP as separator put between PIDs"), fp);
@@ -108,7 +114,7 @@ static char *get_basename (char *filename)
 
 static char *pid_link (pid_t pid, const char *base_name)
 {
-	char link [PROCPATHLEN];
+	char link [1000];
 	char *result;
 	ssize_t path_alloc_size;
 	ssize_t len;
@@ -136,8 +142,10 @@ static char *pid_link (pid_t pid, const char *base_name)
 
 static void select_procs (void)
 {
-	PROCTAB *ptp;
-	proc_t task;
+	enum pids_item items[] = { PIDS_ID_PID, PIDS_CMD, PIDS_CMDLINE_V };
+	enum rel_items { rel_pid, rel_cmd, rel_cmdline };
+	struct pids_info *info = NULL;
+	struct pids_stack *stack;
 	int match;
 	static int size = 0;
 	char *cmd_arg0, *cmd_arg0base;
@@ -150,15 +158,17 @@ static void select_procs (void)
 	/* get the input base name */
 	program_base = get_basename(program);
 
-	ptp = openproc (PROC_FILLCOM | PROC_FILLSTAT);
+	procps_pids_new(&info, items, 3);
 
 	exe_link = root_link = NULL;
-	memset(&task, 0, sizeof (task));
-	while(readproc(ptp, &task)) {
+	while ((stack = procps_pids_get(info, PIDS_FETCH_TASKS_ONLY))) {
+		char  *p_cmd     = PIDS_VAL(rel_cmd,     str,   stack, info),
+		     **p_cmdline = PIDS_VAL(rel_cmdline, strv,  stack, info);
+		int    tid       = PIDS_VAL(rel_pid,     s_int, stack, info);
 
 		if (opt_rootdir_check) {
 			/* get the /proc/<pid>/root symlink value */
-			root_link = pid_link(task.XXXID, "root");
+			root_link = pid_link(tid, "root");
 			match = !strcmp(pidof_root, root_link);
 			safe_free(root_link);
 
@@ -167,9 +177,9 @@ static void select_procs (void)
 			}
 		}
 
-		if (!is_omitted(task.XXXID) && task.cmdline && *task.cmdline) {
+		if (!is_omitted(tid) && ((p_cmdline && *p_cmdline) || opt_with_workers)) {
 
-			cmd_arg0 = *task.cmdline;
+			cmd_arg0 = (p_cmdline && *p_cmdline) ? *p_cmdline : "\0";
 
 			/* processes starting with '-' are login shells */
 			if (*cmd_arg0 == '-') {
@@ -180,7 +190,7 @@ static void select_procs (void)
 			cmd_arg0base = get_basename(cmd_arg0);
 
 			/* get the /proc/<pid>/exe symlink value */
-			exe_link = pid_link(task.XXXID, "exe");
+			exe_link = pid_link(tid, "exe");
 
 			/* get the exe_link base name */
 			exe_link_base = get_basename(exe_link);
@@ -190,22 +200,22 @@ static void select_procs (void)
 			if (!strcmp(program, cmd_arg0base) ||
 			    !strcmp(program_base, cmd_arg0) ||
 			    !strcmp(program, cmd_arg0) ||
-
+			    (opt_with_workers && !strcmp(program, p_cmd)) ||
 			    !strcmp(program, exe_link_base) ||
 			    !strcmp(program, exe_link))
 			{
 				match = 1;
 
-			} else if (opt_scripts_too && *(task.cmdline+1)) {
+			} else if (opt_scripts_too && p_cmdline && *(p_cmdline+1)) {
 
-				cmd_arg1 = *(task.cmdline+1);
+				cmd_arg1 = *(p_cmdline+1);
 
 				/* get the arg1 base name */
-				cmd_arg1base = get_basename(cmd_arg1);
+                cmd_arg1base = get_basename(cmd_arg1);
 
-				/* if script, then task.cmd = argv1, otherwise task.cmd = argv0 */
-				if (task.cmd &&
-				    !strncmp(task.cmd, cmd_arg1base, strlen(task.cmd)) &&
+				/* if script, then cmd = argv1, otherwise cmd = argv0 */
+				if (p_cmd &&
+				    !strncmp(p_cmd, cmd_arg1base, strlen(p_cmd)) &&
 				    (!strcmp(program, cmd_arg1base) ||
 				    !strcmp(program_base, cmd_arg1) ||
 				    !strcmp(program, cmd_arg1)))
@@ -217,7 +227,7 @@ static void select_procs (void)
              * setproctitle so use the cmdline
              */
             if (!match && strchr(cmd_arg0, ' ')) {
-                match = (strcmp(program, task.cmd)==0);
+                match = (strcmp(program, p_cmd)==0);
             }
 
 			safe_free(exe_link);
@@ -228,16 +238,17 @@ static void select_procs (void)
 					procs = xrealloc(procs, size * (sizeof *procs));
 				}
 				if (procs) {
-					procs[proc_count++].pid = task.XXXID;
+					procs[proc_count++].pid = tid;
 				} else {
 					xerrx(EXIT_FAILURE, _("internal error"));
 				}
 			}
 
 		}
+
 	}
 
-	closeproc (ptp);
+	procps_pids_unref(&info);
 }
 
 
@@ -289,13 +300,15 @@ int main (int argc, char **argv)
 	int first_pid = 1;
 
 	const char *separator = " ";
-	const char *opts = "scnxmo:S:?Vh";
+	const char *opts = "scnqxwmo:S:?Vh";
 
 	static const struct option longopts[] = {
 		{"check-root", no_argument, NULL, 'c'},
 		{"single-shot", no_argument, NULL, 's'},
 		{"omit-pid", required_argument, NULL, 'o'},
-		{"separator", required_argument, NULL, 's'},
+		{"separator", required_argument, NULL, 'S'},
+		{"quiet", no_argument, NULL, 'q'},
+		{"with-workers", no_argument, NULL, 'w'},
 		{"help", no_argument, NULL, 'h'},
 		{"version", no_argument, NULL, 'V'},
 		{NULL, 0, NULL, 0}
@@ -312,6 +325,9 @@ int main (int argc, char **argv)
 	/* process command-line options */
 	while ((opt = getopt_long (argc, argv, opts, longopts, NULL)) != -1) {
 		switch (opt) {
+		case 'q':
+			opt_quiet = 1;
+			/* fallthrough */
 		case 's':
 			opt_single_shot = 1;
 			break;
@@ -321,13 +337,17 @@ int main (int argc, char **argv)
 		case 'x':
 			opt_scripts_too = 1;
 			break;
+		case 'w':
+			opt_with_workers = 1;
+			break;
 		case 'c':
 			if (geteuid() == 0) {
 				opt_rootdir_check = 1;
-				safe_free(pidof_root);
+                safe_free(pidof_root);
 				pidof_root = pid_link(getpid(), "root");
 			}
 			break;
+		case 'd': /* sysv pidof uses this for S */
 		case 'S':
 			separator = optarg;
 			break;
@@ -354,17 +374,21 @@ int main (int argc, char **argv)
 
 		program = argv[optind++];
 
+		if (*program == '\0') continue;
+
 		select_procs();	/* get the list of matching processes */
 
 		if (proc_count) {
 
 			found = 1;
 			for (i = proc_count - 1; i >= 0; i--) {	/* and display their PIDs */
-				if (first_pid) {
-					first_pid = 0;
-					printf ("%ld", (long) procs[i].pid);
-				} else {
-					printf ("%s%ld", separator, (long) procs[i].pid);
+				if (!opt_quiet) {
+					if (first_pid) {
+						first_pid = 0;
+						printf ("%ld", (long) procs[i].pid);
+					} else {
+						printf ("%s%ld", separator, (long) procs[i].pid);
+					}
 				}
 				if (opt_single_shot) break;
 			}
@@ -374,7 +398,7 @@ int main (int argc, char **argv)
 	}
 
 	/* final line feed */
-	if (found) printf("\n");
+	if (!opt_quiet && found) printf("\n");
 
 	/* some cleaning */
 	safe_free(procs);
