@@ -30,14 +30,21 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <assert.h>
+#include <limits.h>
 
 #include "c.h"
 #include "fileutils.h"
 #include "nls.h"
-#include "proc/escape.h"
 #include "xalloc.h"
-#include "proc/readproc.h"
-#include "proc/version.h"
+
+#include <proc/pids.h>
+
+static struct pids_info *Pids_info;
+
+enum pids_item Pid_items[] = {
+    PIDS_ID_PID,  PIDS_ID_TGID,
+    PIDS_CMDLINE, PIDS_ADDR_STACK_START };
+enum rel_items { pid, tgid, cmdline, start_stack };
 
 const char *nls_Address,
 	   *nls_Offset,
@@ -87,7 +94,7 @@ static int justify_print(const char *str, int width, int right)
 	return width;
 }
 
-static int integer_width(unsigned KLONG number)
+static int integer_width(unsigned long number)
 {
 	int result;
 
@@ -129,19 +136,18 @@ usage(FILE * out)
 }
 
 static char mapbuf[1024];
-static char cmdbuf[512];
 
-static unsigned KLONG range_low;
-static unsigned KLONG range_high = ~0ull;
+static unsigned long range_low;
+static unsigned long range_high = ~0ul;
 
-static int c_option;
-static int C_option;
-static int d_option;
-static int n_option;
-static int N_option;
-static int q_option;
-static int x_option;
-static int X_option;
+static int c_option = 0;
+static int C_option = 0;
+static int d_option = 0;
+static int n_option = 0;
+static int N_option = 0;
+static int q_option = 0;
+static int x_option = 0;
+static int X_option = 0;
 
 static int map_desc_showpath;
 
@@ -170,12 +176,12 @@ static void discover_shm_minor(void)
 		char perms[32];
 		/* to clean up unprintables */
 		char *tmp;
-		unsigned KLONG start, end;
+		unsigned long start, end;
 		unsigned long long file_offset, inode;
 		unsigned dev_major, dev_minor;
-		sscanf(mapbuf_b, "%" KLF "x-%" KLF "x %31s %llx %x:%x %llu", &start,
-		       &end, perms, &file_offset, &dev_major, &dev_minor,
-		       &inode);
+		if (sscanf(mapbuf_b, "%lx-%lx %31s %llx %x:%x %llu", &start,
+		       &end, perms, &file_offset, &dev_major, &dev_minor, &inode) < 6)
+            continue;
 		tmp = strchr(mapbuf_b, '\n');
 		if (tmp)
 			*tmp = '\0';
@@ -207,12 +213,12 @@ out_destroy:
 	return;
 }
 
-static char *mapping_name(proc_t * p, unsigned KLONG addr,
-				unsigned KLONG len, const char *mapbuf_b,
+static const char *mapping_name(struct pids_stack *p, unsigned long addr,
+				unsigned long len, const char *mapbuf_b,
 				unsigned showpath, unsigned dev_major,
 				unsigned dev_minor, unsigned long long inode)
 {
-	char *cp;
+	const char *cp;
 
 	if (!dev_major && dev_minor == shm_minor && strstr(mapbuf_b, "/SYSV")) {
 		static char shmbuf[64];
@@ -227,16 +233,9 @@ static char *mapping_name(proc_t * p, unsigned KLONG addr,
 		return cp[1] ? cp + 1 : cp;
 	}
 
-	cp = strchr(mapbuf_b, '/');
-	if (cp) {
-		if (showpath)
-			return cp;
-		/* it WILL succeed */
-		return strrchr(cp, '/') + 1;
-	}
-
 	cp = _("  [ anon ]");
-	if ((p->start_stack >= addr) && (p->start_stack <= addr + len))
+	if (PIDS_VAL(start_stack, ul_int, p, Pids_info) >= addr
+	&& (PIDS_VAL(start_stack, ul_int, p, Pids_info) <= addr + len))
 		cp = _("  [ stack ]");
 	return cp;
 }
@@ -246,13 +245,14 @@ static char *mapping_name(proc_t * p, unsigned KLONG addr,
 #define DETL "31"		/* for format strings */
 #define NUM_LENGTH 21		/* python says: len(str(2**64)) == 20 */
 #define NUML "20"		/* for format strings */
-#define VMFLAGS_LENGTH 81 /* There are 27 posible 2 character vmflags as of this patch */
+#define VMFLAGS_LENGTH 128	/* 30 2-char space-separated flags == 90+1, but be safe */
+#define VMFL "127"		/* for format strings */
 
 struct listnode {
 	char description[DETAIL_LENGTH];
 	char value_str[NUM_LENGTH];
-	unsigned KLONG value;
-	unsigned KLONG total;
+	unsigned long value;
+	unsigned long total;
 	int max_width;
 	struct listnode *next;
 };
@@ -326,7 +326,9 @@ static void print_extended_maps (FILE *f)
 		/* If line too long we dump everything else. */
 		c = mapbuf[strlen(mapbuf) - 1];
 		while (c != '\n') {
-			ret = fgets(mapbuf, sizeof mapbuf, f);
+			fgets(mapbuf, sizeof mapbuf, f);
+			if (!ret || !mapbuf[0])
+				xerrx(EXIT_FAILURE, _("Unknown format in smaps file!"));
 			c = mapbuf[strlen(mapbuf) - 1];
 		}
 
@@ -338,8 +340,8 @@ static void print_extended_maps (FILE *f)
 		if (strlen(inode ) > maxw5)	maxw5 = strlen(inode);
 
 		ret = fgets(mapbuf, sizeof mapbuf, f);
-		nfields = sscanf(mapbuf, "%"DETL"[^:]: %"NUML"[0-9] kB %c",
-				 detail_desc, value_str, &c);
+		nfields = ret ? sscanf(mapbuf, "%"DETL"[^:]: %"NUML"[0-9] kB %c",
+					detail_desc, value_str, &c) : 0;
 		listnode = listhead;
 		/* === READ MAPPING DETAILS === */
 		while (ret != NULL && nfields == 2) {
@@ -350,6 +352,9 @@ static void print_extended_maps (FILE *f)
 			if (listnode == NULL) {
 				assert(firstmapping == 2);
 				listnode = calloc(1, sizeof *listnode);
+				if (listnode == NULL)
+					xerrx(EXIT_FAILURE, _("ERROR: memory allocation failed"));
+
 				if (listhead == NULL) {
 					assert(listtail == NULL);
 					listhead = listnode;
@@ -369,7 +374,7 @@ static void print_extended_maps (FILE *f)
 					      mapbuf);
 			}
 			strcpy(listnode->value_str, value_str);
-			sscanf(value_str, "%"KLF"u", &listnode->value);
+			sscanf(value_str, "%lu", &listnode->value);
 			if (firstmapping == 2) {
 				listnode->total += listnode->value;
 				if (q_option) {
@@ -381,16 +386,18 @@ static void print_extended_maps (FILE *f)
 			listnode = listnode->next;
 loop_end:
 			ret = fgets(mapbuf, sizeof mapbuf, f);
-			nfields = sscanf(mapbuf, "%"DETL"[^:]: %"NUML"[0-9] kB %c",
-					 detail_desc, value_str, &c);
+			nfields = ret ? sscanf(mapbuf, "%"DETL"[^:]: %"NUML"[0-9] kB %c",
+						detail_desc, value_str, &c) : 0;
 		}
 
 		/* === GET VMFLAGS === */
-		nfields = sscanf(mapbuf, "VmFlags: %[a-z ]", vmflags);
+		nfields = ret ? sscanf(mapbuf, "VmFlags: %"VMFL"[a-z ]", vmflags) : 0;
 		if (nfields == 1) {
+			int len = strlen(vmflags);
+			if (len > 0 && vmflags[len-1] == ' ') vmflags[--len] = '\0';
+			if (len > maxwv) maxwv = len;
 			if (! has_vmflags) has_vmflags = 1;
 			ret = fgets(mapbuf, sizeof mapbuf, f);
-			if (strlen(vmflags) > maxwv) maxwv = strlen(vmflags);
 		}
 
 		if (firstmapping == 2) { /* width measurement stage, do not print anything yet */
@@ -433,7 +440,7 @@ loop_end:
 					justify_print(listnode->description, listnode->max_width, 1);
 
 				if (has_vmflags && is_enabled("VmFlags"))
-					printf(" %*s", maxwv, "VmFlags");
+					maxwv = justify_print("VmFlags", maxwv, 1);
 
 				if (is_enabled(nls_Mapping))
 					justify_print(nls_Mapping, 0, 0);
@@ -511,14 +518,19 @@ loop_end:
 	/* We don't free() the list, it's used for all PIDs passed as arguments */
 }
 
-static int one_proc(proc_t * p)
+	// variable placed here to silence compiler 'uninitialized' warning
+static unsigned long start_To_Avoid_Warning;
+
+static int one_proc (struct pids_stack *p)
 {
 	char buf[32];
 	FILE *fp;
 	unsigned long total_shared = 0ul;
 	unsigned long total_private_readonly = 0ul;
 	unsigned long total_private_writeable = 0ul;
-	unsigned KLONG diff = 0;
+	unsigned long diff = 0;
+	unsigned long end;
+	char perms[32] = "";
 	const char *cp2 = NULL;
 	unsigned long long rss = 0ull;
 	unsigned long long private_dirty = 0ull;
@@ -528,33 +540,27 @@ static int one_proc(proc_t * p)
 	unsigned long long total_shared_dirty = 0ull;
 	int maxw1=0, maxw2=0, maxw3=0, maxw4=0, maxw5=0;
 
-	/* Overkill, but who knows what is proper? The "w" prog uses
-	 * the tty width to determine this.
-	 */
-	int maxcmd = 0xfffff;
-
-	escape_command(cmdbuf, p, sizeof cmdbuf, &maxcmd,
-		       ESC_ARGS | ESC_BRACKETS);
-	printf("%u:   %s\n", p->tgid, cmdbuf);
+	printf("%u:   %s\n", PIDS_VAL(tgid, s_int, p, Pids_info), PIDS_VAL(cmdline, str, p, Pids_info));
 
 	if (x_option || X_option || c_option) {
-		sprintf(buf, "/proc/%u/smaps", p->tgid);
+		snprintf(buf, sizeof buf, "/proc/%u/smaps", PIDS_VAL(tgid, s_int, p, Pids_info));
 		if ((fp = fopen(buf, "r")) == NULL)
 			return 1;
 	} else {
-		sprintf(buf, "/proc/%u/maps", p->tgid);
+		snprintf(buf, sizeof buf, "/proc/%u/maps", PIDS_VAL(tgid, s_int, p, Pids_info));
 		if ((fp = fopen(buf, "r")) == NULL)
 			return 1;
 	}
 
 	if (X_option || c_option) {
 		print_extended_maps(fp);
+        fclose(fp);
 		return 0;
 	}
 
 	if (x_option) {
 		maxw1 = 16;
-		if (sizeof(KLONG) == 4) maxw1 = 8;
+		if (sizeof(long) == 4) maxw1 = 8;
 		maxw2 = maxw3 = maxw4 = 7;
 		maxw5 = 5;
 		if (!q_option) {
@@ -569,7 +575,7 @@ static int one_proc(proc_t * p)
 
 	if (d_option) {
 		maxw1 = 16;
-		if (sizeof(KLONG) == 4) maxw1 = 8;
+		if (sizeof(long) == 4) maxw1 = 8;
 		maxw2 = 7;
 		maxw3 = 5;
 		maxw4 = 16;
@@ -585,10 +591,8 @@ static int one_proc(proc_t * p)
 	}
 
 	while (fgets(mapbuf, sizeof mapbuf, fp)) {
-		char perms[32];
 		/* to clean up unprintables */
 		char *tmp;
-		unsigned KLONG start, end;
 		unsigned long long file_offset, inode;
 		unsigned dev_major, dev_minor;
 		unsigned long long smap_value;
@@ -614,30 +618,34 @@ static int one_proc(proc_t * p)
 					continue;
 				}
 				if (strcmp("Swap", smap_key) == 0) {
-					/*doesn't matter as long as last */
-					printf("%0*" KLF "x %*lu %*llu %*llu %*s %s\n",
-					       maxw1, start,
-					       maxw2, (unsigned long)(diff >> 10),
-					       maxw3, rss,
-					       maxw4, (private_dirty + shared_dirty),
-					       maxw5, perms,
-					       cp2);
+					/* doesn't matter as long as last */
+					if (cp2)
+                        printf("%0*lx %*lu %*llu %*llu %*s %s\n",
+                               maxw1, start_To_Avoid_Warning,
+                               maxw2, (unsigned long)(diff >> 10),
+                               maxw3, rss,
+                               maxw4, (private_dirty + shared_dirty),
+                               maxw5, perms,
+                               cp2);
 					/* reset some counters */
 					rss = shared_dirty = private_dirty = 0ull;
-					diff = 0;
+                    diff = end = 0;
+                    perms[0] = '\0';
+                    cp2 = NULL;
 					continue;
 				}
 			}
 			/* Other keys or not a key-value pair */
 			continue;
 		}
-		sscanf(mapbuf, "%" KLF "x-%" KLF "x %31s %llx %x:%x %llu", &start,
+		sscanf(mapbuf, "%lx-%lx %31s %llx %x:%x %llu",
+		       &start_To_Avoid_Warning,
 		       &end, perms, &file_offset, &dev_major, &dev_minor,
 		       &inode);
 
 		if (end - 1 < range_low)
 			continue;
-		if (range_high < start)
+		if (range_high < start_To_Avoid_Warning)
 			break;
 
 		tmp = strchr(mapbuf, '\n');
@@ -650,7 +658,7 @@ static int one_proc(proc_t * p)
 			tmp++;
 		}
 
-		diff = end - start;
+		diff = end - start_To_Avoid_Warning;
 		if (perms[3] == 's')
 			total_shared += diff;
 		if (perms[3] == 'p') {
@@ -669,17 +677,17 @@ static int one_proc(proc_t * p)
 
 		if (x_option) {
 			cp2 =
-			    mapping_name(p, start, diff, mapbuf, map_desc_showpath, dev_major,
+			    mapping_name(p, start_To_Avoid_Warning, diff, mapbuf, map_desc_showpath, dev_major,
 					 dev_minor, inode);
 			/* printed with the keys */
 			continue;
 		}
 		if (d_option) {
 			const char *cp =
-			    mapping_name(p, start, diff, mapbuf, map_desc_showpath, dev_major,
+			    mapping_name(p, start_To_Avoid_Warning, diff, mapbuf, map_desc_showpath, dev_major,
 					 dev_minor, inode);
-			printf("%0*" KLF "x %*lu %*s %0*llx %*.*s%03x:%05x %s\n",
-			       maxw1, start,
+			printf("%0*lx %*lu %*s %0*llx %*.*s%03x:%05x %s\n",
+			       maxw1, start_To_Avoid_Warning,
 			       maxw2, (unsigned long)(diff >> 10),
 			       maxw3, perms,
 			       maxw4, file_offset,
@@ -688,19 +696,19 @@ static int one_proc(proc_t * p)
 		}
 		if (!x_option && !d_option) {
 			const char *cp =
-			    mapping_name(p, start, diff, mapbuf, map_desc_showpath, dev_major,
+			    mapping_name(p, start_To_Avoid_Warning, diff, mapbuf, map_desc_showpath, dev_major,
 					 dev_minor, inode);
-			printf((sizeof(KLONG) == 8)
-			       ? "%016" KLF "x %6luK %s %s\n"
+			printf((sizeof(long) == 8)
+			       ? "%016lx %6luK %s %s\n"
 			       : "%08lx %6luK %s %s\n",
-			       start, (unsigned long)(diff >> 10), perms, cp);
+			       start_To_Avoid_Warning, (unsigned long)(diff >> 10), perms, cp);
 		}
 
 	}
 	fclose(fp);
 	if (!q_option) {
 		if (x_option) {
-			if (sizeof(KLONG) == 4)
+			if (sizeof(long) == 4)
 				justify_print("--------", maxw1, 0);
 			else
 				justify_print("----------------", maxw1, 0);
@@ -726,7 +734,7 @@ static int one_proc(proc_t * p)
 			     total_private_writeable >> 10, total_shared >> 10);
 		}
 		if (!x_option && !d_option) {
-			if (sizeof(KLONG) == 8)
+			if (sizeof(long) == 8)
 				/* Translation Hint: keep total string length
 				 * as 24 characters. Adjust %16 if needed*/
 				printf(_(" total %16ldK\n"),
@@ -744,26 +752,30 @@ static int one_proc(proc_t * p)
 	return 0;
 }
 
-static void range_arguments(char *optarg)
+static void range_arguments(const char *optarg)
 {
-	char *arg1;
-	char *arg2;
+	char *buf, *arg1, *arg2;
 
-	arg1 = xstrdup(optarg);
+	if ((buf = xstrdup(optarg)) == NULL) {
+        xerrx(EXIT_FAILURE, "%s: '%s'", _("failed to parse argument"),
+              (optarg?optarg:"(null)"));
+    }
+	arg1 = buf;
 	arg2 = strchr(arg1, ',');
 	if (arg2)
-		*arg2 = '\0';
-	if (arg2)
-		++arg2;
+		*arg2++ = '\0';
 	else
 		arg2 = arg1;
-	if (arg1 && *arg1)
-		range_low = STRTOUKL(arg1, &arg1, 16);
-	if (*arg2)
-		range_high = STRTOUKL(arg2, &arg2, 16);
-	if (arg1 && (*arg1 || *arg2))
+	if (arg1[0] != '\0')
+		range_low = strtoul(arg1, &arg1, 16);
+	if (arg2[0] != '\0')
+		range_high = strtoul(arg2, &arg2, 16);
+	if (*arg1 || *arg2) {
+        free(buf);
 		xerrx(EXIT_FAILURE, "%s: '%s'", _("failed to parse argument"),
 		      optarg);
+    }
+    free(buf);
 }
 
 
@@ -828,7 +840,7 @@ static int config_read (char *rc_filename)
 				trimmed += SECTION_STR_MAPPING_LEN;
 				section_id = SECTION_ID_MAPPING;
 			} else {
-				while (*trimmed != ']' || *trimmed == '\0') trimmed++;
+				while (*trimmed != ']' && *trimmed != '\0') trimmed++;
 				if (*trimmed == ']') {
 					section_id = SECTION_ID_UNSUPPORTED;
 					xwarnx(_("unsupported section found in the config - line %d"), line_cnt);
@@ -897,10 +909,13 @@ static int config_read (char *rc_filename)
 }
 
 
-static int config_create (char *rc_filename)
+static int config_create (const char *rc_filename)
 {
 	FILE *f;
 
+	if (rc_filename == NULL)
+		return 0;
+
 	/* check if rc exists */
 	f = fopen(rc_filename, "r");
 
@@ -989,10 +1004,9 @@ static char *get_default_rc_filename(void)
 
 int main(int argc, char **argv)
 {
-	pid_t *pidlist;
-	unsigned count = 0;
-	PROCTAB *PT;
-	proc_t p;
+	struct pids_fetch *pids_fetch;
+	unsigned *pidlist;
+	int reap_count, user_count;
 	int ret = 0, c, conf_ret;
 	char *rc_filename = NULL;
 
@@ -1114,7 +1128,6 @@ int main(int argc, char **argv)
 	if (C_option) c_option = 1;
 
 	if (c_option) {
-
 		if (!C_option) rc_filename = get_default_rc_filename();
 
 		if (!rc_filename) return(EXIT_FAILURE);
@@ -1130,11 +1143,14 @@ int main(int argc, char **argv)
 				return(EXIT_FAILURE);
 			}
 		}
-
 	}
+	if ((size_t)argc >= INT_MAX / sizeof(pid_t))
+		xerrx(EXIT_FAILURE, _("too many arguments"));
+	if (procps_pids_new(&Pids_info, Pid_items, 4))
+		xerrx(EXIT_FAILURE, _("library failed pids statistics"));
+	pidlist = xmalloc(sizeof(pid_t) * argc);
 
-	pidlist = xmalloc(sizeof(pid_t) * (argc+1));
-
+	user_count = 0;
 	while (*argv) {
 		char *walk = *argv++;
 		char *endp;
@@ -1150,21 +1166,20 @@ int main(int argc, char **argv)
 		pid = strtoul(walk, &endp, 0);
 		if (pid < 1ul || pid > 0x7ffffffful || *endp)
 			usage(stderr);
-		pidlist[count++] = pid;
+		pidlist[user_count++] = pid;
 	}
 
 	discover_shm_minor();
 
-	memset(&p, '\0', sizeof(p));
-	/* old libproc interface is zero-terminated */
-	pidlist[count] = 0;
-	PT = openproc(PROC_FILLSTAT | PROC_FILLARG | PROC_PID, pidlist);
-	while (readproc(PT, &p)) {
-		ret |= one_proc(&p);
-		count--;
+	if (!(pids_fetch = procps_pids_select(Pids_info, pidlist, user_count, PIDS_SELECT_PID)))
+		xerrx(EXIT_FAILURE, _("library failed pids statistics"));
+
+	for (reap_count = 0; reap_count < pids_fetch->counts->total; reap_count++) {
+		ret |= one_proc(pids_fetch->stacks[reap_count]);
 	}
-	closeproc(PT);
+
 	free(pidlist);
+	procps_pids_unref(&Pids_info);
 
 	/* cleaning the list used for the -c/-X/-XX modes */
 	for (listnode = listhead; listnode != NULL ; ) {
@@ -1180,7 +1195,7 @@ int main(int argc, char **argv)
 		cnf_listhead = cnf_listnode;
 	}
 
-	if (count)
+	if (reap_count < user_count)
 		/* didn't find all processes asked for */
 		ret |= 42;
 	return ret;
