@@ -22,11 +22,12 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <limits.h>
 
 #include <sys/types.h>
 
-#include "../proc/readproc.h"
-#include "../proc/sysinfo.h"
+#include "../proc/misc.h"
+#include "../include/xalloc.h"
 
 #include "common.h"
 
@@ -45,9 +46,9 @@ static format_node *do_one_spec(const char *spec, const char *override){
   if(fs){
     int w1, w2;
     format_node *thisnode;
-    thisnode = malloc(sizeof(format_node));
+    thisnode = xmalloc(sizeof(format_node));
     if(fs->flags & CF_PIDMAX){
-      w1 = (int)get_pid_digits();
+      w1 = (int)procps_pid_length();
       w2 = strlen(fs->head);
       if(w2>w1) w1=w2; // FIXME w/ separate header/body column sizing
     }else{
@@ -62,7 +63,6 @@ static format_node *do_one_spec(const char *spec, const char *override){
       thisnode->name = strdup(fs->head);
     }
     thisnode->pr = fs->pr;
-    thisnode->need = fs->need;
     thisnode->vendor = fs->vendor;
     thisnode->flags = fs->flags;
     thisnode->next = NULL;
@@ -172,7 +172,7 @@ static const char *aix_format_parse(sf_node *sfn){
         return _("AIX field descriptor processing bug");
       }
     } else {
-      int len;
+      size_t len;
       len = strcspn(walk, "%");
       memcpy(buf,walk,len);
       if(0){
@@ -182,11 +182,10 @@ double_percent:
       }
       buf[len] = '\0';
       walk += len;
-      fnode = malloc(sizeof(format_node));
-      fnode->width = len;
+      fnode = xmalloc(sizeof(format_node));
+      fnode->width = len < INT_MAX ? len : INT_MAX;
       fnode->name = strdup(buf);
       fnode->pr = NULL;     /* checked for */
-      fnode->need = 0;
       fnode->vendor = AIX;
       fnode->flags = CF_PRINT_EVERY_TIME;
       fnode->next = NULL;
@@ -259,6 +258,7 @@ static const char *format_parse(sf_node *sfn){
     format_node *endp;
     char *equal_loc;
     char *colon_loc;
+    if(!walk) catastrophic_failure(__FILE__, __LINE__, _("please report this bug"));
     sep_loc = strpbrk(walk," ,\t\n");
     /* if items left, then sep_loc is not in header override */
     if(items && sep_loc) *sep_loc = '\0';
@@ -271,7 +271,7 @@ static const char *format_parse(sf_node *sfn){
     if(colon_loc){   /* if width override */
       *colon_loc = '\0';
       colon_loc++;
-      if(strspn(colon_loc,"0123456789") != strlen(colon_loc) || *colon_loc=='0' || !*colon_loc){
+      if(strspn(colon_loc,"0123456789") != strlen(colon_loc) || *colon_loc=='0' || !*colon_loc || atoi(colon_loc) <= 0){
         free(buf);
         goto badwidth;
       }
@@ -296,11 +296,12 @@ static const char *format_parse(sf_node *sfn){
       }
       // FIXME: enforce signal width to 8, 9, or 16 (grep: SIGNAL wide_signals)
       fnode->width = atoi(colon_loc); // already verified to be a number
+      if(fnode->width <= 0) catastrophic_failure(__FILE__, __LINE__, _("please report this bug"));
     }
     endp = fnode; while(endp->next) endp = endp->next;  /* find end */
     endp->next = sfn->f_cooked;
     sfn->f_cooked = fnode;
-    walk = sep_loc + 1; /* point to next item, if any */
+    walk = sep_loc ? sep_loc + 1 : NULL; /* point to next item, if any */
   }
   free(buf);
   already_parsed_format = 1;
@@ -312,16 +313,17 @@ static const char *format_parse(sf_node *sfn){
   if(0) improper: err=_("improper format list");
   if(0) badwidth: err=_("column widths must be unsigned decimal numbers");
   if(0) notmacro: err=_("can not set width for a macro (multi-column) format specifier");
-  if(strchr(sfn->sf,'%')) err = aix_format_parse(sfn);
+  if (!err)
+    if(strchr(sfn->sf,'%')) err = aix_format_parse(sfn);
   return err;
 }
 
 /****************  Parse single sort specifier *******************/
 static sort_node *do_one_sort_spec(const char *spec){
   const format_struct *fs;
-  int reverse = 0;
+  enum pids_sort_order reverse = PIDS_SORT_ASCEND;
   if(*spec == '-'){
-    reverse = 1;
+    reverse = PIDS_SORT_DESCEND;
     spec++;
   } else if(*spec == '+'){
     spec++;
@@ -329,9 +331,10 @@ static sort_node *do_one_sort_spec(const char *spec){
   fs = search_format_array(spec);
   if(fs){
     sort_node *thisnode;
-    thisnode = malloc(sizeof(sort_node));
+    thisnode = xmalloc(sizeof(sort_node));
     thisnode->sr = fs->sr;
-    thisnode->need = fs->need;
+    // next is a special pointer, called to help with rel enums
+    thisnode->xe = (int(*)(char*,proc_t*))fs->pr;
     thisnode->reverse = reverse;
     thisnode->next = NULL;
     return thisnode;
@@ -428,6 +431,7 @@ static const char *verify_short_sort(const char *arg){
   walk = arg;
   for(;;){
     tmp = *walk;
+    if(tmp < 0 || (size_t)tmp >= sizeof(checkoff)) return _("bad sorting code");
     switch(tmp){
     case '\0':
       return NULL;   /* looks good */
@@ -453,7 +457,7 @@ static const char *verify_short_sort(const char *arg){
 
 /************ parse short sorting option *************/
 static const char *short_sort_parse(sf_node *sfn){
-  int direction = 0;
+  enum pids_sort_order direction = PIDS_SORT_ASCEND;
   const char *walk;
   int tmp;
   sort_node *snode;
@@ -467,10 +471,10 @@ static const char *short_sort_parse(sf_node *sfn){
       already_parsed_sort = 1;
       return NULL;
     case '+':
-      direction = 0;
+      direction = PIDS_SORT_ASCEND;
       break;
     case '-':
-      direction = 1;
+      direction = PIDS_SORT_DESCEND;
       break;
     default:
       ss = search_shortsort_array(tmp);
@@ -556,7 +560,7 @@ int defer_sf_option(const char *arg, int source){
   const format_struct *fs;
   int need_item = 1;
 
-  sfn = malloc(sizeof(sf_node));
+  sfn = xmalloc(sizeof(sf_node));
   sfn->sf = strdup(arg);
   sfn->sf_code = source;
   sfn->s_cooked = NULL;
@@ -680,11 +684,10 @@ static const char *generate_sysv_list(void){
   if( (format_flags & FF_Ul) && !(format_modifiers & FM_y) ){
     if(personality & PER_IRIX_l){ /* add "rss" then ':' here */
       PUSH("sgi_rss");
-      fn = malloc(sizeof(format_node));
+      fn = xmalloc(sizeof(format_node));
       fn->width = 1;
       fn->name = strdup(":");
       fn->pr = NULL;     /* checked for */
-      fn->need = 0;
       fn->vendor = AIX;   /* yes, for SGI weirdness */
       fn->flags = CF_PRINT_EVERY_TIME;
       fn->next = format_list;
@@ -777,12 +780,11 @@ const char *process_sf_options(void){
     sort_node *srt_walk;
     srt_walk = sf_walk->s_cooked;
     sf_walk->s_cooked = NULL;
-    while(srt_walk){   /* put any nodes onto sort_list in opposite way */
-      sort_node *travler;
-      travler = srt_walk;
-      srt_walk = srt_walk->next;
+    if (srt_walk) {
+      sort_node *travler = srt_walk;
+      while (travler->next) travler = travler->next;
       travler->next = sort_list;
-      sort_list = travler;
+      sort_list = srt_walk;
     }
     sf_walk = sf_walk->next;
   }
