@@ -17,6 +17,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  *
+ * Part of this code comes from systemd, especially sysctl.c
  * Changelog:
  *            v1.01:
  *                   - added -p <preload> to preload values from a file
@@ -40,13 +41,13 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <ctype.h>
 
 #include "c.h"
 #include "fileutils.h"
 #include "nls.h"
 #include "xalloc.h"
-#include "proc/procps.h"
-#include "proc/version.h"
+#include "procio.h"
 
 /*
  *    Globals...
@@ -64,12 +65,56 @@ static bool PrintName;
 static bool PrintNewline;
 static bool IgnoreError;
 static bool Quiet;
+static bool DryRun;
 static char *pattern;
 
+#define LINELEN 4096
+static char *iobuf;
+static size_t iolen = LINELEN;
+
+typedef struct SysctlSetting {
+    char *key;
+    char *path;
+    char *value;
+    bool ignore_failure;
+    bool glob_exclude;
+    struct SysctlSetting *next;
+} SysctlSetting;
+
+typedef struct SettingList {
+    struct SysctlSetting *head;
+    struct SysctlSetting *tail;
+} SettingList;
+
+#define GLOB_CHARS "*?["
+static inline bool string_is_glob(const char *p)
+{
+    return !!strpbrk(p, GLOB_CHARS);
+}
+
+
 /* Function prototypes. */
 static int pattern_match(const char *string, const char *pat);
 static int DisplayAll(const char *restrict const path);
 
+static inline bool is_proc_path(
+	const char *path)
+{
+    char *resolved_path;
+
+    if ( (resolved_path = realpath(path, NULL)) == NULL)
+	return false;
+
+    if (strncmp(PROC_PATH, resolved_path, strlen(PROC_PATH)) == 0) {
+	free(resolved_path);
+	return true;
+    }
+
+    xwarnx(_("Path is not under %s: %s"), PROC_PATH, path);
+    free(resolved_path);
+    return false;
+}
+
 static void slashdot(char *restrict p, char old, char new)
 {
 	int warned = 1;
@@ -94,6 +139,83 @@ static void slashdot(char *restrict p, char old, char new)
 	}
 }
 
+#if 0  // avoid '-Wunused-function' warning
+static void setting_free(SysctlSetting *s) {
+    if (!s)
+	return;
+
+    free(s->key);
+    free(s->path);
+    free(s->value);
+    free(s);
+}
+#endif
+
+static SysctlSetting *setting_new(
+	const char *key,
+	const char *value,
+	bool ignore_failure,
+    bool glob_exclude) {
+
+    SysctlSetting *s = NULL;
+    char *path = NULL;
+    int proc_len;
+
+    proc_len = strlen(PROC_PATH);
+    /* used to open the file */
+    path = xmalloc(strlen(key) + proc_len + 2);
+    strcpy(path, PROC_PATH);
+    if (key[0] == '-')
+        strcat(path + proc_len, key+1);
+    else
+        strcat(path + proc_len, key);
+    /* change . to / */
+    slashdot(path + proc_len, '.', '/');
+
+    s = xmalloc(sizeof(SysctlSetting));
+
+    *s = (SysctlSetting) {
+        .key = strdup(key),
+        .path = path,
+        .value = value? strdup(value): NULL,
+        .ignore_failure = ignore_failure,
+        .glob_exclude = glob_exclude,
+        .next = NULL,
+    };
+
+    return s;
+}
+
+static void settinglist_add(SettingList *l, SysctlSetting *s) {
+    SysctlSetting *old_tail;
+
+    if (!l)
+        return;
+
+    if (l->head == NULL)
+        l->head = s;
+
+    if (l->tail != NULL) {
+        old_tail = l->tail;
+        old_tail->next = s;
+    }
+    l->tail = s;
+}
+
+static SysctlSetting *settinglist_findpath(const SettingList *l, const char *path) {
+    SysctlSetting *node;
+
+    for (node=l->head; node != NULL; node = node->next) {
+        if (strcmp(node->path, path) == 0)
+            return node;
+    }
+    return NULL;
+}
+
+/* Function prototypes. */
+static int pattern_match(const char *string, const char *pat);
+static int DisplayAll(const char *restrict const path);
+
 /*
  * Display the usage format
  */
@@ -105,24 +227,25 @@ static void __attribute__ ((__noreturn__))
 	      _(" %s [options] [variable[=value] ...]\n"),
 		program_invocation_short_name);
 	fputs(USAGE_OPTIONS, out);
-	fputs(_("  -a, --all            display all variables\n"
-		"  -A                   alias of -a\n"
-		"  -X                   alias of -a\n"
-		"      --deprecated     include deprecated parameters to listing\n"
-		"  -b, --binary         print value without new line\n"
-		"  -e, --ignore         ignore unknown variables errors\n"
-		"  -N, --names          print variable names without values\n"
-		"  -n, --values         print only values of a variables\n"
-		"  -p, --load[=<file>]  read values from file\n"
-		"  -f                   alias of -p\n"
-		"      --system         read values from all system directories\n"
-		"  -r, --pattern <expression>\n"
-		"                       select setting that match expression\n"
-		"  -q, --quiet          do not echo variable set\n"
-		"  -w, --write          enable writing a value to variable\n"
-		"  -o                   does nothing\n"
-		"  -x                   does nothing\n"
-		"  -d                   alias of -h\n"), out);
+	fputs(_("  -a, --all            display all variables\n"), out);
+	fputs(_("  -A                   alias of -a\n"), out);
+	fputs(_("  -X                   alias of -a\n"), out);
+	fputs(_("      --deprecated     include deprecated parameters to listing\n"), out);
+	fputs(_("      --dry-run        Print the key and values but do not write\n"), out);
+	fputs(_("  -b, --binary         print value without new line\n"), out);
+	fputs(_("  -e, --ignore         ignore unknown variables errors\n"), out);
+	fputs(_("  -N, --names          print variable names without values\n"), out);
+	fputs(_("  -n, --values         print only values of the given variable(s)\n"), out);
+	fputs(_("  -p, --load[=<file>]  read values from file\n"), out);
+	fputs(_("  -f                   alias of -p\n"), out);
+	fputs(_("      --system         read values from all system directories\n"), out);
+	fputs(_("  -r, --pattern <expression>\n"
+		"                       select setting that match expression\n"), out);
+	fputs(_("  -q, --quiet          do not echo variable set\n"), out);
+	fputs(_("  -w, --write          enable writing a value to variable\n"), out);
+	fputs(_("  -o                   does nothing\n"), out);
+	fputs(_("  -x                   does nothing\n"), out);
+	fputs(_("  -d                   alias of -h\n"), out);
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
 	fputs(USAGE_VERSION, out);
@@ -131,6 +254,40 @@ static void __attribute__ ((__noreturn__))
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
 
+/*
+ * Strip left/leading side of a string
+ */
+static char *lstrip(char *line)
+{
+    char *start;
+
+    if (!line || !*line)
+        return line;
+
+    start = line;
+    while(isspace(*start)) start++;
+
+    return start;
+}
+
+/*
+ * Strip right/trailing side of a string
+ * by placing a \0
+ */
+static void rstrip(char *line)
+{
+    char *end;
+
+    if (!line || !*line)
+        return;
+
+    end = line + strlen(line) - 1;
+    while(end > line && isspace(*end)) end--;
+
+    end[1] = '\0';
+}
+
+#if 0  // avoid '-Wunused-function' warning
 /*
  * Strip the leading and trailing spaces from a string
  */
@@ -154,16 +311,17 @@ static char *StripLeadingAndTrailingSpaces(char *oneline)
 
 	return t;
 }
+#endif
 
 /*
  * Read a sysctl setting
  */
 static int ReadSetting(const char *restrict const name)
 {
-	int rc = 0;
+	int rc = EXIT_SUCCESS;
 	char *restrict tmpname;
 	char *restrict outname;
-	char inbuf[1025];
+	ssize_t rlen;
 	FILE *restrict fp;
 	struct stat ts;
 
@@ -192,13 +350,18 @@ static int ReadSetting(const char *restrict const name)
 	if (stat(tmpname, &ts) < 0) {
 		if (!IgnoreError) {
 			xwarn(_("cannot stat %s"), tmpname);
-			rc = -1;
+			rc = EXIT_FAILURE;
 		}
 		goto out;
 	}
 	if ((ts.st_mode & S_IRUSR) == 0)
 		goto out;
 
+	if (!is_proc_path(tmpname)) {
+	    rc = -1;
+	    goto out;
+	}
+
 	if (S_ISDIR(ts.st_mode)) {
 		size_t len;
 		len = strlen(tmpname);
@@ -209,62 +372,71 @@ static int ReadSetting(const char *restrict const name)
 	}
 
 	if (pattern && !pattern_match(outname, pattern)) {
-		free(outname);
-		return 0;
+		rc = EXIT_SUCCESS;
+		goto out;
 	}
 
-	fp = fopen(tmpname, "r");
+	if (NameOnly) {
+		fprintf(stdout, "%s\n", outname);
+		goto out;
+	}
+
+	fp = fprocopen(tmpname, "r");
 
 	if (!fp) {
 		switch (errno) {
 		case ENOENT:
 			if (!IgnoreError) {
 				xwarnx(_("\"%s\" is an unknown key"), outname);
-				rc = -1;
+				rc = EXIT_FAILURE;
 			}
 			break;
 		case EACCES:
 			xwarnx(_("permission denied on key '%s'"), outname);
-			rc = -1;
+			rc = EXIT_FAILURE;
+			break;
+		case EIO:	    /* Ignore stable_secret below /proc/sys/net/ipv6/conf */
+			rc = EXIT_FAILURE;
 			break;
 		default:
 			xwarn(_("reading key \"%s\""), outname);
-			rc = -1;
+			rc = EXIT_FAILURE;
 			break;
 		}
 	} else {
 		errno = 0;
-		if (fgets(inbuf, sizeof inbuf - 1, fp)) {
+		if ((rlen = getline(&iobuf, &iolen, fp)) > 0) {
 			/* this loop is required, see
 			 * /sbin/sysctl -a | egrep -6 dev.cdrom.info
 			 */
 			do {
-				if (NameOnly) {
-					fprintf(stdout, "%s\n", outname);
+				char *nlptr;
+				if (PrintName) {
+					fprintf(stdout, "%s = ", outname);
+					do {
+						fprintf(stdout, "%s", iobuf);
+						nlptr = &iobuf[strlen(iobuf) - 1];
+						/* already has the \n in it */
+						if (*nlptr == '\n')
+							break;
+					} while ((rlen = getline(&iobuf, &iolen, fp)) > 0);
+					if (*nlptr != '\n')
+						putchar('\n');
 				} else {
-					/* already has the \n in it */
-					if (PrintName) {
-						fprintf(stdout, "%s = %s",
-							outname, inbuf);
-						if (inbuf[strlen(inbuf) - 1] != '\n')
-							putchar('\n');
-					} else {
-						if (!PrintNewline) {
-							char *nlptr =
-							    strchr(inbuf, '\n');
-							if (nlptr)
-								*nlptr = '\0';
-						}
-						fprintf(stdout, "%s", inbuf);
+					if (!PrintNewline) {
+						nlptr = strchr(iobuf, '\n');
+						if (nlptr)
+							*nlptr = '\0';
 					}
+					fprintf(stdout, "%s", iobuf);
 				}
-			} while (fgets(inbuf, sizeof inbuf - 1, fp));
+			} while ((rlen = getline(&iobuf, &iolen, fp)) > 0);
 		} else {
 			switch (errno) {
 			case EACCES:
 				xwarnx(_("permission denied on key '%s'"),
 				       outname);
-				rc = -1;
+				rc = EXIT_FAILURE;
 				break;
 			case EISDIR: {
 					size_t len;
@@ -275,9 +447,12 @@ static int ReadSetting(const char *restrict const name)
 					rc = DisplayAll(tmpname);
 					goto out;
 				}
+			case EIO:	    /* Ignore stable_secret below /proc/sys/net/ipv6/conf */
+				rc = EXIT_FAILURE;
+				break;
 			default:
 				xwarnx(_("reading key \"%s\""), outname);
-				rc = -1;
+				rc = EXIT_FAILURE;
 			case 0:
 				break;
 			}
@@ -305,7 +480,7 @@ static int is_deprecated(char *filename)
  */
 static int DisplayAll(const char *restrict const path)
 {
-	int rc = 0;
+	int rc = EXIT_SUCCESS;
 	int rc2;
 	DIR *restrict dp;
 	struct dirent *restrict de;
@@ -315,7 +490,7 @@ static int DisplayAll(const char *restrict const path)
 
 	if (!dp) {
 		xwarnx(_("unable to open directory \"%s\""), path);
-		rc = -1;
+		rc = EXIT_FAILURE;
 	} else {
 		readdir(dp);	/* skip .  */
 		readdir(dp);	/* skip .. */
@@ -351,122 +526,185 @@ static int DisplayAll(const char *restrict const path)
 /*
  * Write a sysctl setting
  */
-static int WriteSetting(const char *setting)
+static int WriteSetting(
+    const char *key,
+    const char *path,
+    const char *value,
+    const bool ignore_failure)
 {
-	int rc = 0;
-	const char *name = setting;
-	const char *value;
-	const char *equals;
-	char *tmpname;
-	char *outname;
-	char *last_dot;
+    int rc = EXIT_SUCCESS;
+    FILE *fp;
+    struct stat ts;
 
-	FILE *fp;
-	struct stat ts;
+    if (!key || !path)
+        return rc;
 
-	if (!name)
-		/* probably don't want to display this err */
-		return 0;
+    if (stat(path, &ts) < 0) {
+        if (!IgnoreError) {
+            xwarn(_("cannot stat %s"), path);
+            rc = EXIT_FAILURE;
+        }
+        return rc;
+    }
 
-	equals = strchr(setting, '=');
+    if (!is_proc_path(path)) {
+        return EXIT_FAILURE;
+    }
 
-	if (!equals) {
-		xwarnx(_("\"%s\" must be of the form name=value"),
-		       setting);
-		return -1;
-	}
+    if ((ts.st_mode & S_IWUSR) == 0) {
+        xwarn(_("setting key \"%s\""), key);
+        return rc;
+    }
 
-	/* point to the value in name=value */
-	value = equals + 1;
+    if (S_ISDIR(ts.st_mode)) {
+        xwarn(_("setting key \"%s\""), key);
+        return rc;
+    }
 
-	if (!*name || !*value || name == equals) {
-		xwarnx(_("malformed setting \"%s\""), setting);
-		return -2;
-	}
+    if (!DryRun) {
+        if ((fp = fprocopen(path, "w")) == NULL) {
+            switch (errno) {
+            case ENOENT:
+                if (!IgnoreError) {
+                    xwarnx(_("\"%s\" is an unknown key%s"),
+                           key, (ignore_failure?_(", ignoring"):""));
+                    if (!ignore_failure)
+                        rc = EXIT_FAILURE;
+                }
+                break;
+            case EPERM:
+            case EROFS:
+            case EACCES:
+                xwarnx(_("permission denied on key \"%s\"%s"),
+                       key, (ignore_failure?_(", ignoring"):""));
+                break;
+            default:
+                xwarn(_("setting key \"%s\"%s"),
+                      key, (ignore_failure?_(", ignoring"):""));
+                break;
+            }
+            if (!ignore_failure && errno != ENOENT)
+                rc = EXIT_FAILURE;
+        } else {
+            if (0 < fprintf(fp, "%s\n", value))
+                rc = EXIT_SUCCESS;
+            if (close_stream(fp) != 0) {
+                xwarn(_("setting key \"%s\""), path);
+                return EXIT_FAILURE;
+            }
+        }
+    }
+    if ((rc == EXIT_SUCCESS && !Quiet) || DryRun) {
+        if (NameOnly) {
+            printf("%s\n", value);
+        } else {
+            if (PrintName) {
+                printf("%s = %s\n", path, value);
+            } else {
+                if (PrintNewline)
+                    printf("%s\n", value);
+                else
+                    printf("%s", value);
+            }
+        }
+    }
+    return rc;
+}
 
-	/* used to open the file */
-	tmpname = xmalloc(equals - name + 1 + strlen(PROC_PATH));
-	strcpy(tmpname, PROC_PATH);
-	strncat(tmpname, name, (int) (equals - name));
-	tmpname[equals - name + strlen(PROC_PATH)] = 0;
-	/* change . to / */
-	slashdot(tmpname + strlen(PROC_PATH), '.', '/');
+/*
+ * parse each configuration line, there are multiple ways of specifying
+ * a key/value here:
+ *
+ * key = value                               simple setting
+ * -key = value                              ignore errors
+ * key.pattern.*.with.glob = value           set keys that match glob
+ * -key.pattern.exclude.with.glob            dont set this value
+ * key.pattern.override.with.glob = value    set this glob match to value
+ *
+ */
 
-	/* used to display the output */
-	outname = xmalloc(equals - name + 1);
-	strncpy(outname, name, (int) (equals - name));
-	outname[equals - name] = 0;
-	/* change / to . */
-	slashdot(outname, '/', '.');
-	last_dot = strrchr(outname, '.');
-	if (last_dot != NULL && is_deprecated(last_dot + 1)) {
-		xwarnx(_("%s is deprecated, value not set"), outname);
-		goto out;
+static SysctlSetting *parse_setting_line(
+    const char *path,
+    const int linenum,
+    char *line)
+{
+    char *key;
+    char *value;
+    bool glob_exclude = FALSE;
+    bool ignore_failure = FALSE;
+
+    key = lstrip(line);
+    if (strlen(key) < 2)
+        return NULL;
+
+    /* skip over comments */
+    if (key[0] == '#' || key[0] == ';')
+        return NULL;
+
+    if (pattern && !pattern_match(key, pattern))
+        return NULL;
+
+    value = strchr(key, '=');
+    if (value == NULL) {
+        if (key[0] == '-') {
+            glob_exclude = TRUE;
+            key++;
+            value = NULL;
+            rstrip(key);
+        } else {
+            xwarnx(_("%s(%d): invalid syntax, continuing..."),
+                   path, linenum);
+            return NULL;
+        }
+    } else {
+        value[0]='\0';
+        if (key[0] == '-') {
+            ignore_failure = TRUE;
+            key++;
+        }
+        value++; // skip over =
+        value=lstrip(value);
+        rstrip(value);
+        rstrip(key);
+    }
+    return setting_new(key, value, ignore_failure, glob_exclude);
+}
+
+/* Go through the setting list, expand and sort out
+ * setting globs and actually write the settings out
+ */
+static int write_setting_list(const SettingList *sl)
+{
+    SysctlSetting *node;
+    int rc = EXIT_SUCCESS;
+
+    for (node=sl->head; node != NULL; node=node->next) {
+        if (node->glob_exclude)
+            continue;
+
+        if (string_is_glob(node->path)) {
+            glob_t globbuf;
+            int i;
+
+            if (glob(node->path, 0, NULL, &globbuf) != 0)
+                continue;
+
+            for(i=0; i < globbuf.gl_pathc; i++) {
+                if (settinglist_findpath(sl, globbuf.gl_pathv[i]))
+                    continue; // override or exclude
+
+                rc |= WriteSetting(node->key, globbuf.gl_pathv[i], node->value,
+                                   node->ignore_failure);
+            }
+        } else {
+            rc |= WriteSetting(node->key, node->path, node->value,
+                               node->ignore_failure);
         }
 
-	if (stat(tmpname, &ts) < 0) {
-		if (!IgnoreError) {
-			xwarn(_("cannot stat %s"), tmpname);
-			rc = -1;
-		}
-		goto out;
-	}
 
-	if ((ts.st_mode & S_IWUSR) == 0) {
-		xwarn(_("setting key \"%s\""), outname);
-		goto out;
-	}
+    }
 
-	if (S_ISDIR(ts.st_mode)) {
-		xwarn(_("setting key \"%s\""), outname);
-		goto out;
-	}
-
-	fp = fopen(tmpname, "w");
-
-	if (!fp) {
-		switch (errno) {
-		case ENOENT:
-			if (!IgnoreError) {
-				xwarnx(_("\"%s\" is an unknown key"), outname);
-				rc = -1;
-			}
-			break;
-		case EACCES:
-			xwarnx(_("permission denied on key '%s'"), outname);
-			rc = -1;
-			break;
-		default:
-			xwarn(_("setting key \"%s\""), outname);
-			rc = -1;
-			break;
-		}
-	} else {
-		rc = fprintf(fp, "%s\n", value);
-		if (0 < rc)
-			rc = 0;
-		if (close_stream(fp) != 0)
-			xwarn(_("setting key \"%s\""), outname);
-		if (rc == 0 && !Quiet) {
-			if (NameOnly) {
-				fprintf(stdout, "%s\n", outname);
-			} else {
-				if (PrintName) {
-					fprintf(stdout, "%s = %s\n",
-						outname, value);
-				} else {
-					if (PrintNewline)
-						fprintf(stdout, "%s\n", value);
-					else
-						fprintf(stdout, "%s", value);
-				}
-			}
-		}
-	}
-      out:
-	free(tmpname);
-	free(outname);
-	return rc;
+    return rc;
 }
 
 static int pattern_match(const char *string, const char *pat)
@@ -487,20 +725,30 @@ static int pattern_match(const char *string, const char *pat)
  * Preload the sysctl's from the conf file.  We parse the file and then
  * reform it (strip out whitespace).
  */
-static int Preload(const char *restrict const filename)
+static int Preload(SettingList *setlist, const char *restrict const filename)
 {
-	char oneline[256];
-	char buffer[256];
 	FILE *fp;
-	char *t;
 	int n = 0;
-	int rc = 0;
-	char *name, *value;
+	int rc = EXIT_SUCCESS;
+	ssize_t rlen;
 	glob_t globbuf;
 	int globerr;
+	int globflg;
 	int j;
 
-	globerr = glob(filename, GLOB_NOCHECK | GLOB_TILDE, NULL, &globbuf);
+	globflg = GLOB_NOCHECK;
+#ifdef GLOB_BRACE
+	globflg |= GLOB_BRACE;
+#endif
+#ifdef GLOB_TILDE
+	globflg |= GLOB_TILDE;
+#else
+	if (filename[0] == '~')
+		xwarnx(_("GLOB_TILDE is not supported on your platform, "
+			 "the tilde in \"%s\" won't be expanded."), filename);
+#endif
+	globerr = glob(filename, globflg, NULL, &globbuf);
+
 	if (globerr != 0 && globerr != GLOB_NOMATCH)
 		xerr(EXIT_FAILURE, _("glob failed"));
 
@@ -509,44 +757,21 @@ static int Preload(const char *restrict const filename)
 		    ? stdin : fopen(globbuf.gl_pathv[j], "r");
 		if (!fp) {
 			xwarn(_("cannot open \"%s\""), globbuf.gl_pathv[j]);
-			return -1;
+            return EXIT_FAILURE;
 		}
 
-		while (fgets(oneline, sizeof oneline, fp)) {
+		while ((rlen =  getline(&iobuf, &iolen, fp)) > 0) {
+            SysctlSetting *setting;
+
 			n++;
-			t = StripLeadingAndTrailingSpaces(oneline);
 
-			if (strlen(t) < 2)
+			if (rlen < 2)
 				continue;
 
-			if (*t == '#' || *t == ';')
-				continue;
-
-			name = strtok(t, "=");
-			if (!name || !*name) {
-				xwarnx(_("%s(%d): invalid syntax, continuing..."),
-				       globbuf.gl_pathv[j], n);
-				continue;
-			}
-
-			StripLeadingAndTrailingSpaces(name);
-
-			if (pattern && !pattern_match(name, pattern))
-				continue;
-
-			value = strtok(NULL, "\n\r");
-			if (!value || !*value) {
-				xwarnx(_("%s(%d): invalid syntax, continuing..."),
-				       globbuf.gl_pathv[j], n);
-				continue;
-			}
-
-			while ((*value == ' ' || *value == '\t') && *value != 0)
-				value++;
-
-			/* should NameOnly affect this? */
-			sprintf(buffer, "%s=%s", name, value);
-			rc |= WriteSetting(buffer);
+            if ( (setting = parse_setting_line(globbuf.gl_pathv[j], n, iobuf))
+                 == NULL)
+                continue;
+            settinglist_add(setlist, setting);
 		}
 
 		fclose(fp);
@@ -566,18 +791,20 @@ static int sortpairs(const void *A, const void *B)
 	return strcmp(a->name, b->name);
 }
 
-static int PreloadSystem(void)
+static int PreloadSystem(SettingList *setlist)
 {
 	unsigned di, i;
 	const char *dirs[] = {
-		"/run/sysctl.d",
 		"/etc/sysctl.d",
+		"/run/sysctl.d",
 		"/usr/local/lib/sysctl.d",
 		"/usr/lib/sysctl.d",
 		"/lib/sysctl.d",
 	};
 	struct pair **cfgs = NULL;
 	unsigned ncfgs = 0;
+	int rc = EXIT_SUCCESS;
+	struct stat ts;
 	enum { nprealloc = 16 };
 
 	for (di = 0; di < sizeof(dirs) / sizeof(dirs[0]); ++di) {
@@ -634,12 +861,23 @@ static int PreloadSystem(void)
 	for (i = 0; i < ncfgs; ++i) {
 		if (!Quiet)
 			printf(_("* Applying %s ...\n"), cfgs[i]->value);
-		Preload(cfgs[i]->value);
+		rc |= Preload(setlist, cfgs[i]->value);
 	}
 
-	if (!Quiet)
-		printf(_("* Applying %s ...\n"), DEFAULT_PRELOAD);
-	return Preload(DEFAULT_PRELOAD);
+
+	if (stat(DEFAULT_PRELOAD, &ts) == 0 && S_ISREG(ts.st_mode)) {
+		if (!Quiet)
+			printf(_("* Applying %s ...\n"), DEFAULT_PRELOAD);
+		rc |= Preload(setlist, DEFAULT_PRELOAD);
+	}
+
+	/* cleaning */
+	for (i = 0; i < ncfgs; ++i) {
+		free(cfgs[i]);
+	}
+	if (cfgs) free(cfgs);
+
+	return rc;
 }
 
 /*
@@ -652,15 +890,19 @@ int main(int argc, char *argv[])
 	bool preloadfileOpt = false;
 	int ReturnCode = 0;
 	int c;
+	int rc = 0;
 	const char *preloadfile = NULL;
+    SettingList *setlist;
 
 	enum {
 		DEPRECATED_OPTION = CHAR_MAX + 1,
-		SYSTEM_OPTION
+		SYSTEM_OPTION,
+        DRYRUN_OPTION
 	};
 	static const struct option longopts[] = {
 		{"all", no_argument, NULL, 'a'},
 		{"deprecated", no_argument, NULL, DEPRECATED_OPTION},
+		{"dry-run", no_argument, NULL, DRYRUN_OPTION},
 		{"binary", no_argument, NULL, 'b'},
 		{"ignore", no_argument, NULL, 'e'},
 		{"names", no_argument, NULL, 'N'},
@@ -688,6 +930,10 @@ int main(int argc, char *argv[])
 	IgnoreError = false;
 	Quiet = false;
 	IgnoreDeprecated = true;
+    DryRun = false;
+    setlist = xmalloc(sizeof(SettingList));
+    setlist->head = NULL;
+    setlist->tail = NULL;
 
 	if (argc < 2)
 		Usage(stderr);
@@ -740,7 +986,12 @@ int main(int argc, char *argv[])
 			break;
 		case SYSTEM_OPTION:
 			IgnoreError = true;
-			return PreloadSystem();
+			rc |= PreloadSystem(setlist);
+            rc |= write_setting_list(setlist);
+            return rc;
+        case DRYRUN_OPTION:
+            DryRun = true;
+            break;
 		case 'r':
 			pattern = xstrdup(optarg);
 			break;
@@ -759,6 +1010,8 @@ int main(int argc, char *argv[])
 	argc -= optind;
 	argv += optind;
 
+	iobuf = xmalloc(iolen);
+
 	if (DisplayAllOpt)
 		return DisplayAll(PROC_PATH);
 
@@ -766,15 +1019,16 @@ int main(int argc, char *argv[])
 		int ret = EXIT_SUCCESS, i;
 		if (!preloadfile) {
 			if (!argc) {
-				ret |= Preload(DEFAULT_PRELOAD);
+				ret |= Preload(setlist, DEFAULT_PRELOAD);
 			}
 		} else {
 			/* This happens when -pfile option is
 			 * used without space. */
-			ret |= Preload(preloadfile);
+			ret |= Preload(setlist, preloadfile);
 		}
 		for (i = 0; i < argc; i++)
-			ret |= Preload(argv[i]);
+			ret |= Preload(setlist, argv[i]);
+        ret |= write_setting_list(setlist);
 		return ret;
 	}
 
@@ -788,9 +1042,14 @@ int main(int argc, char *argv[])
 		      program_invocation_short_name);
 
 	for ( ; *argv; argv++) {
-		if (WriteMode || index(*argv, '='))
-			ReturnCode += WriteSetting(*argv);
-		else
+		if (WriteMode || strchr(*argv, '=')) {
+            SysctlSetting *s;
+            if ( (s = parse_setting_line("command line", 0, *argv)) != NULL)
+                ReturnCode |= WriteSetting(s->key, s->path, s->value,
+                                           s->ignore_failure);
+            else
+                ReturnCode |= EXIT_FAILURE;
+        } else
 			ReturnCode += ReadSetting(*argv);
 	}
 	return ReturnCode;
