@@ -23,17 +23,6 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include "c.h"
-#include "fileutils.h"
-#include "nls.h"
-#include "proc/devname.h"
-#include "proc/escape.h"
-#include "proc/procps.h"
-#include "proc/readproc.h"
-#include "proc/sysinfo.h"
-#include "proc/version.h"
-#include "proc/whattime.h"
-
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -54,14 +43,34 @@
 #include <termios.h>
 #include <time.h>
 #include <unistd.h>
-#include <utmp.h>
+#ifdef HAVE_UTMPX_H
+#	include <utmpx.h>
+#else
+#	include <utmp.h>
+#endif
 #include <arpa/inet.h>
 
+#include "c.h"
+#include "fileutils.h"
+#include "nls.h"
+
+#include <proc/misc.h>
+#include <proc/pids.h>
+
 static int ignoreuser = 0;	/* for '-u' */
 static int oldstyle = 0;	/* for '-o' */
-static proc_t **procs;		/* our snapshot of the process table */
 
+#ifdef HAVE_UTMPX_H
+typedef struct utmpx utmp_t;
+#else
 typedef struct utmp utmp_t;
+#endif
+
+#if !defined(UT_HOSTSIZE) || defined(__UT_HOSTSIZE)
+#	define UT_HOSTSIZE __UT_HOSTSIZE
+#	define UT_LINESIZE __UT_LINESIZE
+#	define UT_NAMESIZE __UT_NAMESIZE
+#endif
 
 #ifdef W_SHOWFROM
 # define FROM_STRING "on"
@@ -113,21 +122,22 @@ static void print_host(const char *restrict host, int len, const int fromlen)
 /* This routine prints the display part of the host or IPv6 link address interface */
 static void print_display_or_interface(const char *restrict host, int len, int restlen)
 {
-	char *disp,*tmp;
+	const char *const end = host + (len > 0 ? len : 0);
+	const char *disp, *tmp;
 
 	if (restlen <= 0) return; /* not enough space for printing anything */
 
 	/* search for a collon (might be a display) */
-	disp = (char *)host;
-	while ( (disp < (host + len)) && (*disp != ':') && isprint(*disp) ) disp++;
+	disp = host;
+	while ( (disp < end) && (*disp != ':') && isprint(*disp) ) disp++;
 
 	/* colon found */
-	if (*disp == ':') {
+	if (disp < end && *disp == ':') {
 		/* detect multiple colons -> IPv6 in the host (not a display) */
 		tmp = disp+1;
-		while ( (tmp < (host + len)) && (*tmp != ':') && isprint(*tmp) ) tmp++;
+		while ( (tmp < end) && (*tmp != ':') && isprint(*tmp) ) tmp++;
 
-		if (*tmp != ':') { /* multiple colons not found - it's a display */
+		if (tmp >= end || *tmp != ':') { /* multiple colons not found - it's a display */
 
 			/* number of chars till the end of the input field */
 			len -= (disp - host);
@@ -149,9 +159,9 @@ static void print_display_or_interface(const char *restrict host, int len, int r
 		} else { /* multiple colons found - it's an IPv6 address */
 
 			/* search for % (interface separator in case of IPv6 link address) */
-			while ( (tmp < (host + len)) && (*tmp != '%') && isprint(*tmp) ) tmp++;
+			while ( (tmp < end) && (*tmp != '%') && isprint(*tmp) ) tmp++;
 
-			if (*tmp == '%') { /* interface separator found */
+			if (tmp < end && *tmp == '%') { /* interface separator found */
 
 				/* number of chars till the end of the input field */
 				len -= (tmp - host);
@@ -170,7 +180,6 @@ static void print_display_or_interface(const char *restrict host, int len, int r
 					fputc('-', stdout);
 				}
 			}
-
 		}
 	}
 
@@ -187,6 +196,7 @@ static void print_from(const utmp_t *restrict const u, const int ip_addresses, c
 	char buf[fromlen + 1];
 	char buf_ipv6[INET6_ADDRSTRLEN];
 	int len;
+#ifndef __CYGWIN__
 	int32_t ut_addr_v6[4];      /* IP address of the remote host */
 
 	if (ip_addresses) { /* -i switch used */
@@ -224,6 +234,9 @@ static void print_from(const utmp_t *restrict const u, const int ip_addresses, c
 	} else {  /* -i switch NOT used */
 		print_host(u->ut_host, UT_HOSTSIZE, fromlen);
 	}
+#else
+	print_host(u->ut_host, UT_HOSTSIZE, fromlen);
+#endif
 }
 
 
@@ -311,6 +324,27 @@ static void print_logintime(time_t logt, FILE * fout)
 	}
 }
 
+/*
+ * Get the Device ID of the given TTY
+ */
+static int get_tty_device(const char *restrict const name)
+{
+    struct stat st;
+    static char buf[32];
+    char *dev_paths[] = { "/dev/%s", "/dev/tty%s", "/dev/pts/%s", NULL};
+    int i;
+
+    if (name[0] == '/' && stat(name, &st) == 0)
+        return st.st_rdev;
+
+    for (i=0; dev_paths[i] != NULL; i++) {
+        snprintf(buf, 32, dev_paths[i], name);
+        if (stat(buf, &st) == 0)
+            return st.st_rdev;
+    }
+    return -1;
+}
+
 /*
  * This function scans the process table accumulating total cpu
  * times for any processes "associated" with this login session.
@@ -318,127 +352,168 @@ static void print_logintime(time_t logt, FILE * fout)
  * the user for that login session is doing currently. This the
  * essential core of 'w'.
  */
-static const proc_t *getproc(const utmp_t * restrict const u,
-			     const char *restrict const tty,
-			     unsigned long long *restrict const jcpu,
-			     int *restrict const found_utpid)
+static int find_best_proc(
+        const utmp_t * restrict const u,
+        const char *restrict const tty,
+        unsigned long long *restrict const jcpu,
+        unsigned long long *restrict const pcpu,
+        char *cmdline)
 {
-	int line;
-	proc_t **pptr = procs;
-	const proc_t *best = NULL;
-	const proc_t *secondbest = NULL;
-	unsigned uid = ~0U;
+#define PIDS_GETINT(e) PIDS_VAL(EU_ ## e, s_int, reap->stacks[i], info)
+#define PIDS_GETUNT(e) PIDS_VAL(EU_ ## e, u_int, reap->stacks[i], info)
+#define PIDS_GETULL(e) PIDS_VAL(EU_ ## e, ull_int, reap->stacks[i], info)
+#define PIDS_GETSTR(e) PIDS_VAL(EU_ ## e, str, reap->stacks[i], info)
+    unsigned uid = ~0U;
+    int found_utpid = 0;
+    int i, total_procs, line;
+    unsigned long long best_time = 0;
+    unsigned long long secondbest_time = 0;
 
-	*found_utpid = 0;
-	if (!ignoreuser) {
-		char buf[UT_NAMESIZE + 1];
-		/* pointer to static data */
-		struct passwd *passwd_data;
-		strncpy(buf, u->ut_user, UT_NAMESIZE);
-		buf[UT_NAMESIZE] = '\0';
-		passwd_data = getpwnam(buf);
-		if (!passwd_data)
-			return NULL;
-		uid = passwd_data->pw_uid;
-		/* OK to have passwd_data go out of scope here */
-	}
-	line = tty_to_dev(tty);
-	*jcpu = 0;
-	for (; *pptr; pptr++) {
-		const proc_t *restrict const tmp = *pptr;
-		if (unlikely(tmp->tgid == u->ut_pid)) {
-			*found_utpid = 1;
-            if (!best)
-                best = tmp;
-		}
-		if (tmp->tty != line)
-			continue;
-		(*jcpu) += tmp->utime + tmp->stime;
-		secondbest = tmp;
-		/* same time-logic here as for "best" below */
-		if (!(secondbest && tmp->start_time <= secondbest->start_time)) {
-			secondbest = tmp;
-		}
-		if (!ignoreuser && uid != tmp->euid && uid != tmp->ruid)
-			continue;
-		if (tmp->pgrp != tmp->tpgid)
-			continue;
-		if (best && tmp->start_time <= best->start_time)
-			continue;
-		best = tmp;
-	}
-	return best ? best : secondbest;
+    struct pids_info *info=NULL;
+    struct pids_fetch *reap;
+    enum pids_item items[] = {
+        PIDS_ID_TGID,
+        PIDS_TICS_BEGAN,
+        PIDS_ID_EUID,
+        PIDS_ID_RUID,
+        PIDS_ID_TPGID,
+        PIDS_ID_PGRP,
+        PIDS_TTY,
+        PIDS_TICS_ALL,
+        PIDS_CMDLINE};
+    enum rel_items {
+        EU_TGID, EU_START, EU_EUID, EU_RUID, EU_TPGID, EU_PGRP, EU_TTY,
+        EU_TICS_ALL, EU_CMDLINE};
+
+    *jcpu = 0;
+    *pcpu = 0;
+    if (!ignoreuser) {
+        char buf[UT_NAMESIZE + 1];
+        struct passwd *passwd_data;
+        strncpy(buf, u->ut_user, UT_NAMESIZE);
+        buf[UT_NAMESIZE] = '\0';
+        if ((passwd_data = getpwnam(buf)) == NULL)
+            return 0;
+        uid = passwd_data->pw_uid;
+        /* OK to have passwd_data go out of scope here */
+    }
+
+    line = get_tty_device(tty);
+
+    if (procps_pids_new(&info, items, 9) < 0)
+        xerrx(EXIT_FAILURE,
+              _("Unable to create pid info structure"));
+    if ((reap = procps_pids_reap(info, PIDS_FETCH_TASKS_ONLY)) == NULL)
+        xerrx(EXIT_FAILURE,
+              _("Unable to load process information"));
+    total_procs = reap->counts->total;
+
+    for (i=0; i < total_procs; i++) {
+        /* is this the login process? */
+        if (PIDS_GETINT(TGID) == u->ut_pid) {
+            found_utpid = 1;
+            if (!best_time) {
+                best_time = PIDS_GETULL(START);
+                strncpy(cmdline, PIDS_GETSTR(CMDLINE), MAX_CMD_WIDTH);
+                *pcpu = PIDS_GETULL(TICS_ALL);
+            }
+
+        }
+        if (PIDS_GETINT(TTY) != line)
+            continue;
+        (*jcpu) += PIDS_VAL(EU_TICS_ALL, ull_int, reap->stacks[i], info);
+        if (!(secondbest_time && PIDS_GETULL(START) <= secondbest_time)) {
+            secondbest_time = PIDS_GETULL(START);
+            if (cmdline[0] == '-' && cmdline[1] == '\0') {
+                strncpy(cmdline, PIDS_GETSTR(CMDLINE), MAX_CMD_WIDTH);
+                *pcpu = PIDS_GETULL(TICS_ALL);
+            }
+        }
+        if (
+            (!ignoreuser && uid != PIDS_GETUNT(EUID)
+             && uid != PIDS_GETUNT(RUID))
+            || (PIDS_GETINT(PGRP) != PIDS_GETINT(TPGID))
+            || (PIDS_GETULL(START) <= best_time)
+           )
+            continue;
+        best_time = PIDS_GETULL(START);
+        strncpy(cmdline, PIDS_GETSTR(CMDLINE), MAX_CMD_WIDTH);
+        *pcpu = PIDS_GETULL(TICS_ALL);
+    }
+    procps_pids_unref(&info);
+    return found_utpid;
+#undef PIDS_GETINT
+#undef PIDS_GETUNT
+#undef PIDS_GETULL
+#undef PIDS_GETSTR
 }
 
-static void showinfo(utmp_t * u, int formtype, int maxcmd, int from,
-		     const int userlen, const int fromlen, const int ip_addresses)
+static void showinfo(
+            utmp_t * u, int formtype, int maxcmd, int from,
+            const int userlen, const int fromlen, const int ip_addresses)
 {
-	unsigned long long jcpu;
-	int ut_pid_found;
-	unsigned i;
-	char uname[UT_NAMESIZE + 1] = "", tty[5 + UT_LINESIZE + 1] = "/dev/";
-	const proc_t *best;
+    unsigned long long jcpu, pcpu;
+    unsigned i;
+    char uname[UT_NAMESIZE + 1] = "", tty[5 + UT_LINESIZE + 1] = "/dev/";
+    long hertz;
+    char cmdline[MAX_CMD_WIDTH + 1];
 
-	for (i = 0; i < UT_LINESIZE; i++)
-		/* clean up tty if garbled */
-		if (isalnum(u->ut_line[i]) || (u->ut_line[i] == '/'))
-			tty[i + 5] = u->ut_line[i];
-		else
-			tty[i + 5] = '\0';
+    strcpy(cmdline, "-");
 
-	best = getproc(u, tty + 5, &jcpu, &ut_pid_found);
+    hertz = procps_hertz_get();
+    for (i = 0; i < UT_LINESIZE; i++)
+        /* clean up tty if garbled */
+        if (isalnum(u->ut_line[i]) || (u->ut_line[i] == '/'))
+            tty[i + 5] = u->ut_line[i];
+        else
+            tty[i + 5] = '\0';
 
-	/*
-	 * just skip if stale utmp entry (i.e. login proc doesn't
-	 * exist). If there is a desire a cmdline flag could be
-	 * added to optionally show it with a prefix of (stale)
-	 * in front of cmd or something like that.
-	 */
-	if (!ut_pid_found)
-		return;
+    if (find_best_proc(u, tty + 5, &jcpu, &pcpu, cmdline) == 0)
+    /*
+     * just skip if stale utmp entry (i.e. login proc doesn't
+     * exist). If there is a desire a cmdline flag could be
+     * added to optionally show it with a prefix of (stale)
+     * in front of cmd or something like that.
+     */
+        return;
 
-	/* force NUL term for printf */
-	strncpy(uname, u->ut_user, UT_NAMESIZE);
+    /* force NUL term for printf */
+    strncpy(uname, u->ut_user, UT_NAMESIZE);
 
-	if (formtype) {
-		printf("%-*.*s%-9.8s", userlen + 1, userlen, uname, u->ut_line);
-		if (from)
-			print_from(u, ip_addresses, fromlen);
-		print_logintime(u->ut_time, stdout);
-		if (*u->ut_line == ':')
-			/* idle unknown for xdm logins */
-			printf(" ?xdm? ");
-		else
-			print_time_ival7(idletime(tty), 0, stdout);
-		print_time_ival7(jcpu / Hertz, (jcpu % Hertz) * (100. / Hertz),
-				 stdout);
-		if (best) {
-			unsigned long long pcpu = best->utime + best->stime;
-			print_time_ival7(pcpu / Hertz,
-					 (pcpu % Hertz) * (100. / Hertz),
-					 stdout);
-		} else
-			printf("   ?   ");
-	} else {
-		printf("%-*.*s%-9.8s", userlen + 1, userlen, u->ut_user,
-		       u->ut_line);
-		if (from)
-			print_from(u, ip_addresses, fromlen);
-		if (*u->ut_line == ':')
-			/* idle unknown for xdm logins */
-			printf(" ?xdm? ");
-		else
-			print_time_ival7(idletime(tty), 0, stdout);
-	}
-	fputs(" ", stdout);
-	if (likely(best)) {
-		char cmdbuf[MAX_CMD_WIDTH];
-		escape_command(cmdbuf, best, sizeof cmdbuf, &maxcmd, ESC_ARGS);
-		fputs(cmdbuf, stdout);
-	} else {
-		printf("-");
-	}
-	fputc('\n', stdout);
+    if (formtype) {
+        printf("%-*.*s%-9.8s", userlen + 1, userlen, uname, u->ut_line);
+        if (from)
+            print_from(u, ip_addresses, fromlen);
+#ifdef HAVE_UTMPX_H
+        print_logintime(u->ut_tv.tv_sec, stdout);
+#else
+        print_logintime(u->ut_time, stdout);
+#endif
+        if (*u->ut_line == ':')
+            /* idle unknown for xdm logins */
+            printf(" ?xdm? ");
+        else
+            print_time_ival7(idletime(tty), 0, stdout);
+        print_time_ival7(jcpu / hertz, (jcpu % hertz) * (100. / hertz),
+                 stdout);
+        if (pcpu > 0)
+            print_time_ival7(pcpu / hertz,
+                             (pcpu % hertz) * (100. / hertz),
+                             stdout);
+        else
+            printf("   ?   ");
+    } else {
+        printf("%-*.*s%-9.8s", userlen + 1, userlen, u->ut_user,
+               u->ut_line);
+        if (from)
+            print_from(u, ip_addresses, fromlen);
+        if (*u->ut_line == ':')
+            /* idle unknown for xdm logins */
+            printf(" ?xdm? ");
+        else
+            print_time_ival7(idletime(tty), 0, stdout);
+    }
+    printf(" %.*s\n", maxcmd, cmdline);
 }
 
 static void __attribute__ ((__noreturn__))
@@ -446,7 +521,7 @@ static void __attribute__ ((__noreturn__))
 {
 	fputs(USAGE_HEADER, out);
 	fprintf(out,
-              _(" %s [options]\n"), program_invocation_short_name);
+              _(" %s [options] [user]\n"), program_invocation_short_name);
 	fputs(USAGE_OPTIONS, out);
 	fputs(_(" -h, --no-header     do not print header\n"),out);
 	fputs(_(" -u, --no-current    ignore current process username\n"),out);
@@ -513,9 +588,6 @@ int main(int argc, char **argv)
 		case 'h':
 			header = 0;
 			break;
-		case 'l':
-			longform = 1;
-			break;
 		case 's':
 			longform = 0;
 			break;
@@ -571,17 +643,19 @@ int main(int argc, char **argv)
 		maxcmd = atoi(p);
 	else
 		maxcmd = MAX_CMD_WIDTH;
-	if (MAX_CMD_WIDTH < maxcmd)
-		maxcmd = MAX_CMD_WIDTH;
+#define CLAMP_CMD_WIDTH(cw) do { \
+	if ((cw) < MIN_CMD_WIDTH) (cw) = MIN_CMD_WIDTH; \
+	if ((cw) > MAX_CMD_WIDTH) (cw) = MAX_CMD_WIDTH; \
+} while (0)
+	CLAMP_CMD_WIDTH(maxcmd);
 	maxcmd -= 21 + userlen + (from ? fromlen : 0) + (longform ? 20 : 0);
-	if (maxcmd < MIN_CMD_WIDTH)
-        maxcmd = MIN_CMD_WIDTH;
+	CLAMP_CMD_WIDTH(maxcmd);
+#undef CLAMP_CMD_WIDTH
 
-	procs = readproctab(PROC_FILLCOM | PROC_FILLUSR | PROC_FILLSTAT);
 
 	if (header) {
 		/* print uptime and headers */
-		print_uptime(0);
+		printf("%s\n", procps_uptime_sprint());
 		/* Translation Hint: Following five uppercase messages are
 		 * headers. Try to keep alignment intact.  */
 		printf(_("%-*s TTY      "), userlen, _("USER"));
@@ -593,12 +667,20 @@ int main(int argc, char **argv)
 			printf(_("   IDLE WHAT\n"));
 	}
 
+#ifdef HAVE_UTMPX_H
+	setutxent();
+#else
 	utmpname(UTMP_FILE);
 	setutent();
+#endif
 	if (user) {
 		for (;;) {
+#ifdef HAVE_UTMPX_H
+			u = getutxent();
+#else
 			u = getutent();
-			if (unlikely(!u))
+#endif
+			if (!u)
 				break;
 			if (u->ut_type != USER_PROCESS)
 				continue;
@@ -608,8 +690,12 @@ int main(int argc, char **argv)
 		}
 	} else {
 		for (;;) {
+#ifdef HAVE_UTMPX_H
+			u = getutxent();
+#else
 			u = getutent();
-			if (unlikely(!u))
+#endif
+			if (!u)
 				break;
 			if (u->ut_type != USER_PROCESS)
 				continue;
@@ -618,7 +704,11 @@ int main(int argc, char **argv)
 					 fromlen, ip_addresses);
 		}
 	}
+#ifdef HAVE_UTMPX_H
+	endutxent();
+#else
 	endutent();
+#endif
 
 	return EXIT_SUCCESS;
 }
