@@ -78,6 +78,7 @@ static int screen_size_changed = 0;
 static int first_screen = 1;
 static int show_title = 2;	/* number of lines used, 2 or 0 */
 static int precise_timekeeping = 0;
+static int line_wrap = 1;
 
 #define min(x,y) ((x) > (y) ? (y) : (x))
 #define MAX_ANSIBUF 100
@@ -98,6 +99,7 @@ static void __attribute__ ((__noreturn__))
 	fputs(_("  -n, --interval <secs>  seconds to wait between updates\n"), out);
 	fputs(_("  -p, --precise          attempt run command in precise intervals\n"), out);
 	fputs(_("  -t, --no-title         turn off header\n"), out);
+	fputs(_("  -w, --no-wrap          turn off line wrapping\n"), out);
 	fputs(_("  -x, --exec             pass command to exec instead of \"sh -c\"\n"), out);
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
@@ -111,6 +113,7 @@ static int nr_of_colors;
 static int attributes;
 static int fg_col;
 static int bg_col;
+static int more_colors;
 
 
 static void reset_ansi(void)
@@ -126,17 +129,78 @@ static void init_ansi_colors(void)
 		-1, COLOR_BLACK, COLOR_RED, COLOR_GREEN, COLOR_YELLOW,
 		COLOR_BLUE, COLOR_MAGENTA, COLOR_CYAN, COLOR_WHITE
 	};
-
 	nr_of_colors = sizeof(ncurses_colors) / sizeof(short);
 
+	more_colors = (COLORS >= 16) && (COLOR_PAIRS >= 16 * 16);
+
+	// Initialize ncurses colors. -1 is terminal default
+	// 0-7 are auto created standard colors initialized by ncurses
+	if (more_colors) {
+		// Initialize using ANSI SGR 8-bit specified colors
+		// 8-15 are bright colors
+		init_color(8, 333, 333, 333);  // Bright black
+		init_color(9, 1000, 333, 333);  // Bright red
+		init_color(10, 333, 1000, 333);  // Bright green
+		init_color(11, 1000, 1000, 333);  // Bright yellow
+		init_color(12, 333, 333, 1000);  // Bright blue
+		init_color(13, 1000, 333, 1000);  // Bright magenta
+		init_color(14, 333, 1000, 1000);  // Bright cyan
+		// Often ncurses is built with only 256 colors, so lets
+		// stop here - so we can support the -1 terminal default
+		//init_color(15, 1000, 1000, 1000);  // Bright white
+		nr_of_colors += 7;
+	}
+
+	// Initialize all color pairs with ncurses
 	for (bg_col = 0; bg_col < nr_of_colors; bg_col++)
 		for (fg_col = 0; fg_col < nr_of_colors; fg_col++)
-			init_pair(bg_col * nr_of_colors + fg_col + 1, ncurses_colors[fg_col], ncurses_colors[bg_col]);
+			init_pair(bg_col * nr_of_colors + fg_col + 1, fg_col - 1, bg_col - 1);
+
 	reset_ansi();
 }
 
 
-static void set_ansi_attribute(const int attrib)
+static int process_ansi_color_escape_sequence(char** escape_sequence) {
+	// process SGR ANSI color escape sequence
+	// Eg 8-bit
+	// 38;5;⟨n⟩  (set fg color to n)
+	// 48;5;⟨n⟩  (set bg color to n)
+	//
+	// Eg 24-bit (not yet implemented)
+	// ESC[ 38;2;⟨r⟩;⟨g⟩;⟨b⟩ m Select RGB foreground color
+	// ESC[ 48;2;⟨r⟩;⟨g⟩;⟨b⟩ m Select RGB background color
+	int num;
+
+	if (!escape_sequence)
+		return 0; /* avoid NULLPTR dereference, return "not understood" */
+
+	if ((*escape_sequence)[0] != ';')
+		return 0; /* not understood */
+
+	if ((*escape_sequence)[1] == '5') {
+		// 8 bit! ANSI specifies a predefined set of 256 colors here.
+		if ((*escape_sequence)[2] != ';')
+			return 0; /* not understood */
+		num = strtol((*escape_sequence) + 3, escape_sequence, 10);
+		if (num >= 0 && num <= 7) {
+			// 0-7 are standard colors  same as SGR 30-37
+			return num + 1;
+		}
+		if (num >= 8 && num <= 15) {
+			// 8-15 are standard colors  same as SGR 90-97
+			 return more_colors ? num + 1 : num - 8 + 1;
+		}
+
+		// Remainder aren't yet implemented
+		// 16-231:  6 × 6 × 6 cube (216 colors): 16 + 36 × r + 6 × g + b (0 ≤ r, g, b ≤ 5)
+		// 232-255:  grayscale from black to white in 24 steps
+	}
+
+	return 0; /* not understood */
+}
+
+
+static int set_ansi_attribute(const int attrib, char** escape_sequence)
 {
 	switch (attrib) {
 	case -1:	/* restore last settings */
@@ -184,24 +248,54 @@ static void set_ansi_attribute(const int attrib)
 	case 27:	/* unset inversed */
 		attributes &= ~A_REVERSE;
 		break;
+    case 38:
+        fg_col = process_ansi_color_escape_sequence(escape_sequence);
+        if (fg_col == 0) {
+            return 0; /* not understood */
+        }
+        break;
+	case 39:
+		fg_col = 0;
+		break;
+    case 48:
+        bg_col = process_ansi_color_escape_sequence(escape_sequence);
+        if (bg_col == 0) {
+            return 0; /* not understood */
+        }
+        break;
+    case 49:
+        bg_col = 0;
+        break;
 	default:
 		if (attrib >= 30 && attrib <= 37) {	/* set foreground color */
 			fg_col = attrib - 30 + 1;
 		} else if (attrib >= 40 && attrib <= 47) { /* set background color */
 			bg_col = attrib - 40 + 1;
+		} else if (attrib >= 90 && attrib <= 97) { /* set bright fg color */
+			fg_col = more_colors ? attrib - 90 + 9 : attrib - 90 + 1;
+		} else if (attrib >= 100 && attrib <= 107) { /* set bright bg color */
+			bg_col = more_colors ? attrib - 100 + 9 : attrib - 100 + 1;
+		} else {
+			return 0; /* Not understood */
 		}
 	}
-
-	attrset(attributes | COLOR_PAIR(bg_col * nr_of_colors + fg_col + 1));
+    attr_set(attributes, bg_col * nr_of_colors + fg_col + 1, NULL);
+    return 1;
 }
 
 static void process_ansi(FILE * fp)
 {
 	int i, c;
 	char buf[MAX_ANSIBUF];
-	char *numstart, *endptr;
+	char *numstart, *endptr = buf;
+    int ansi_attribute;
 
 	c = getc(fp);
+
+	if (c == '(') {
+		c = getc(fp);
+		c = getc(fp);
+	}
 	if (c != '[') {
 		ungetc(c, fp);
 		return;
@@ -213,9 +307,7 @@ static void process_ansi(FILE * fp)
 			buf[i] = '\0';
 			break;
 		}
-		if (c < '0' && c > '9' && c != ';') {
-			while (--i >= 0)
-				ungetc(buf[i], fp);
+		if ((c < '0' || c > '9') && c != ';') {
 			return;
 		}
 		buf[i] = (char)c;
@@ -228,10 +320,17 @@ static void process_ansi(FILE * fp)
 	 * attributes to apply, but typically there are between 1 and 3.
 	 */
 
-	if (*endptr == '\0') set_ansi_attribute(0); /* [m treated as [0m */
+    /* Special case of <ESC>[m */
+    if (buf[0] == '\0')
+        set_ansi_attribute(0, NULL);
 
-	for (endptr = numstart = buf; *endptr != '\0'; numstart = endptr + 1)
-		set_ansi_attribute(strtol(numstart, &endptr, 10));
+    for (endptr = numstart = buf; *endptr != '\0'; numstart = endptr + 1) {
+        ansi_attribute = strtol(numstart, &endptr, 10);
+        if (!set_ansi_attribute(ansi_attribute, &endptr))
+            break;
+        if (numstart == endptr)
+            set_ansi_attribute(0, NULL); /* [m treated as [0m */
+    }
 }
 
 static void __attribute__ ((__noreturn__)) do_exit(int status)
@@ -359,69 +458,101 @@ wint_t my_getwc(FILE * s)
 #endif	/* WITH_WATCH8BIT */
 
 #ifdef WITH_WATCH8BIT
-static void output_header(wchar_t *restrict wcommand, int wcommand_columns, int wcommand_characters, double interval)
+static void output_header(wchar_t *restrict wcommand, int wcommand_characters, double interval)
 #else
 static void output_header(char *restrict command, double interval)
 #endif	/* WITH_WATCH8BIT */
 {
 	time_t t = time(NULL);
 	char *ts = ctime(&t);
-	int tsl = strlen(ts);
 	char *header;
+	char *right_header;
+	int max_host_name_len = (int) sysconf(_SC_HOST_NAME_MAX);
+	char hostname[max_host_name_len + 1];
+	int command_columns = 0;	/* not including final \0 */
+
+	gethostname(hostname, sizeof(hostname));
 
 	/*
-	 * left justify interval and command, right justify time,
+	 * left justify interval and command, right justify hostname and time,
 	 * clipping all to fit window width
 	 */
 	int hlen = asprintf(&header, _("Every %.1fs: "), interval);
+	int rhlen = asprintf(&right_header, _("%s: %s"), hostname, ts);
 
 	/*
 	 * the rules:
-	 *   width < tsl : print nothing
-	 *   width < tsl + hlen + 1: print ts
-	 *   width = tsl + hlen + 1: print header, ts
-	 *   width < tsl + hlen + 4: print header, ..., ts
-	 *   width < tsl + hlen +    wcommand_columns: print header,
-	 *                           truncated wcommand, ..., ts
-	 *   width > "": print header, wcomand, ts
+	 *   width < rhlen : print nothing
+	 *   width < rhlen + hlen + 1: print hostname, ts
+	 *   width = rhlen + hlen + 1: print header, hostname, ts
+	 *   width < rhlen + hlen + 4: print header, ..., hostname, ts
+	 *   width < rhlen + hlen + wcommand_columns: print header,
+	 *                           truncated wcommand, ..., hostname, ts
+	 *   width > "": print header, wcomand, hostname, ts
 	 * this is slightly different from how it used to be
 	 */
-	if (width < tsl) {
+	if (width < rhlen) {
 		free(header);
+		free(right_header);
 		return;
 	}
-	if (tsl + hlen + 1 <= width) {
+	if (rhlen + hlen + 1 <= width) {
 		mvaddstr(0, 0, header);
-		if (tsl + hlen + 2 <= width) {
-			if (width < tsl + hlen + 4) {
-				mvaddstr(0, width - tsl - 4, "... ");
+		if (rhlen + hlen + 2 <= width) {
+			if (width < rhlen + hlen + 4) {
+				mvaddstr(0, width - rhlen - 4, "... ");
 			} else {
 #ifdef WITH_WATCH8BIT
-				if (width < tsl + hlen + wcommand_columns) {
+	            command_columns = wcswidth(wcommand, -1);
+				if (width < rhlen + hlen + command_columns) {
 					/* print truncated */
-					int available = width - tsl - hlen;
-					int in_use = wcommand_columns;
+					int available = width - rhlen - hlen;
+					int in_use = command_columns;
 					int wcomm_len = wcommand_characters;
 					while (available - 4 < in_use) {
 						wcomm_len--;
 						in_use = wcswidth(wcommand, wcomm_len);
 					}
 					mvaddnwstr(0, hlen, wcommand, wcomm_len);
-					mvaddstr(0, width - tsl - 4, "... ");
+					mvaddstr(0, width - rhlen - 4, "... ");
 				} else {
 					mvaddwstr(0, hlen, wcommand);
 				}
 #else
-				mvaddnstr(0, hlen, command, width - tsl - hlen);
+                command_columns = strlen(command);
+                if (width < rhlen + hlen + command_columns) {
+                    /* print truncated */
+                    mvaddnstr(0, hlen, command, width - rhlen - hlen - 4);
+                    mvaddstr(0, width - rhlen - 4, "... ");
+                } else {
+                    mvaddnstr(0, hlen, command, width - rhlen - hlen);
+                }
 #endif	/* WITH_WATCH8BIT */
 			}
 		}
 	}
-	mvaddstr(0, width - tsl + 1, ts);
+	mvaddstr(0, width - rhlen + 1, right_header);
 	free(header);
+	free(right_header);
 	return;
 }
 
+static void find_eol(FILE *p)
+{
+    int c;
+#ifdef WITH_WATCH8BIT
+    do {
+	c = my_getwc(p);
+    } while (c != WEOF
+	    && c!= L'\n');
+#else
+    do {
+	c = getc(p);
+    } while (c != EOF
+	    && c != '\n');
+#endif /* WITH_WATCH8BIT */
+}
+
 static int run_command(char *restrict command, char **restrict command_argv)
 {
 	FILE *p;
@@ -479,7 +610,7 @@ static int run_command(char *restrict command, char **restrict command_argv)
 	for (y = show_title; y < height; y++) {
 		int eolseen = 0, tabpending = 0, tabwaspending = 0;
 		if (flags & WATCH_COLOR)
-			set_ansi_attribute(-1);
+			set_ansi_attribute(-1, NULL);
 #ifdef WITH_WATCH8BIT
 		wint_t carry = WEOF;
 #endif
@@ -492,7 +623,7 @@ static int run_command(char *restrict command, char **restrict command_argv)
 			int attr = 0;
 
 			if (tabwaspending && (flags & WATCH_COLOR))
-				set_ansi_attribute(-1);
+				set_ansi_attribute(-1, NULL);
 			tabwaspending = 0;
 
 			if (!eolseen) {
@@ -612,6 +743,12 @@ static int run_command(char *restrict command, char **restrict command_argv)
 #endif
 		}
 		oldeolseen = eolseen;
+		if (!line_wrap) {
+		    reset_ansi();
+		    if (flags & WATCH_COLOR)
+			attrset(A_NORMAL);
+		    find_eol(p);
+		}
 	}
 
 	fclose(p);
@@ -643,6 +780,7 @@ int main(int argc, char *argv[])
 {
 	int optc;
 	double interval = 2;
+	char *interval_string;
 	char *command;
 	char **command_argv;
 	int command_length = 0;	/* not including final \0 */
@@ -650,7 +788,6 @@ int main(int argc, char *argv[])
 				 * keeping only */
 #ifdef WITH_WATCH8BIT
 	wchar_t *wcommand = NULL;
-	int wcommand_columns = 0;	/* not including final \0 */
 	int wcommand_characters = 0;	/* not including final \0 */
 #endif	/* WITH_WATCH8BIT */
 
@@ -665,6 +802,7 @@ int main(int argc, char *argv[])
 		{"exec", no_argument, 0, 'x'},
 		{"precise", no_argument, 0, 'p'},
 		{"no-title", no_argument, 0, 't'},
+		{"no-wrap", no_argument, 0, 'w'},
 		{"version", no_argument, 0, 'v'},
 		{0, 0, 0, 0}
 	};
@@ -677,8 +815,12 @@ int main(int argc, char *argv[])
 	textdomain(PACKAGE);
 	atexit(close_stdout);
 
+	interval_string = getenv("WATCH_INTERVAL");
+	if(interval_string != NULL)
+		interval = strtod_nol_or_err(interval_string, _("Could not parse interval from WATCH_INTERVAL"));
+
 	while ((optc =
-		getopt_long(argc, argv, "+bced::ghn:pvtx", longopts, (int *)0))
+		getopt_long(argc, argv, "+bced::ghn:pvtwx", longopts, (int *)0))
 	       != EOF) {
 		switch (optc) {
 		case 'b':
@@ -701,15 +843,14 @@ int main(int argc, char *argv[])
 		case 't':
 			show_title = 0;
 			break;
+		case 'w':
+			line_wrap = 0;
+			break;
 		case 'x':
 			flags |= WATCH_EXEC;
 			break;
 		case 'n':
-			interval = strtod_or_err(optarg, _("failed to parse argument"));
-			if (interval < 0.1)
-				interval = 0.1;
-			if (interval > UINT_MAX)
-				interval = UINT_MAX;
+			interval = strtod_nol_or_err(optarg, _("failed to parse argument"));
 			break;
 		case 'p':
 			precise_timekeeping = 1;
@@ -726,6 +867,11 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (interval < 0.1)
+		interval = 0.1;
+	if (interval > UINT_MAX)
+		interval = UINT_MAX;
+
 	if (optind >= argc)
 		usage(stderr);
 
@@ -762,7 +908,6 @@ int main(int argc, char *argv[])
 		exit(EXIT_FAILURE);
 	}
 	mbstowcs(wcommand, command, wcommand_characters + 1);
-	wcommand_columns = wcswidth(wcommand, -1);
 #endif	/* WITH_WATCH8BIT */
 
 	get_terminal_size();
@@ -783,7 +928,7 @@ int main(int argc, char *argv[])
 			use_default_colors();
 			init_ansi_colors();
 		} else {
-			flags |= WATCH_COLOR;
+			flags &= ~WATCH_COLOR;
 		}
 	}
 	nonl();
@@ -805,7 +950,7 @@ int main(int argc, char *argv[])
 
 		if (show_title)
 #ifdef WITH_WATCH8BIT
-			output_header(wcommand, wcommand_columns, wcommand_characters, interval);
+			output_header(wcommand, wcommand_characters, interval);
 #else
 			output_header(command, interval);
 #endif	/* WITH_WATCH8BIT */
